__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"

import io
import os

from octoprint import UMASK
from octoprint.util import atomic_write


class AbstractFileWrapper:
    """
    Wrapper for file representations to save to storages.

    Arguments:
        filename (str): The file's name
    """

    DEFAULT_PERMISSIONS = 0o664

    def __init__(self, filename):
        self.filename = filename

    def save(self, path, permissions=None):
        """
        Saves the file's content to the given absolute path.

        Arguments:
            path (str): The absolute path to where to save the file
            permissions (int): The permissions to set on the file
        """
        raise NotImplementedError()

    def stream(self):
        """
        Returns a Python stream object (subclass of io.IOBase) representing the file's contents.

        Returns:
            io.IOBase: The file's contents as a stream.
        """
        raise NotImplementedError()


class DiskFileWrapper(AbstractFileWrapper):
    """
    An implementation of :class:`.AbstractFileWrapper` that wraps an actual file on disk. The `save` implementations
    will either copy the file to the new path (preserving file attributes) or -- if `move` is `True` (the default) --
    move the file.

    Arguments:
        filename (str): The file's name
        path (str): The file's absolute path
        move (boolean): Whether to move the file upon saving (True, default) or copying.
    """

    def __init__(self, filename, path, move=True):
        AbstractFileWrapper.__init__(self, filename)
        self.path = path
        self.move = move

    def save(self, path, permissions=None):
        import shutil

        if self.move:
            shutil.move(self.path, path)
        else:
            shutil.copy2(self.path, path)

        if permissions is None:
            permissions = self.DEFAULT_PERMISSIONS & ~UMASK
        os.chmod(path, permissions)

    def stream(self):
        return open(self.path, "rb")


class StreamWrapper(AbstractFileWrapper):
    """
    A wrapper allowing processing of one or more consecutive streams.

    Arguments:
        *streams: One or more :py:class:`io.IOBase` streams to process one after another to save to storage.
    """

    def __init__(self, filename, *streams):
        if not len(streams) > 0:
            raise ValueError("Need at least one stream to wrap")

        AbstractFileWrapper.__init__(self, filename)
        self.streams = streams

    def save(self, path, permissions=None):
        """
        Will dump the contents of all streams provided during construction into the target file, in the order they were
        provided.
        """
        import shutil

        with atomic_write(path, mode="wb") as dest:
            for source in self.streams:
                shutil.copyfileobj(source, dest)
        if permissions is None:
            permissions = self.DEFAULT_PERMISSIONS & ~UMASK
        os.chmod(path, permissions)

    def stream(self):
        """
        If more than one stream was provided to the constructor, will return a :class:`.MultiStream` wrapping all
        provided streams in the order they were provided, else the first and only stream is returned directly.
        """
        if len(self.streams) > 1:
            return MultiStream(*self.streams)
        else:
            return self.streams[0]


class MultiStream(io.RawIOBase):
    """
    A stream implementation which when read reads from multiple streams, one after the other, basically concatenating
    their contents in the order they are provided to the constructor.

    Arguments:
        *streams: One or more :py:class:`io.IOBase` streams to concatenate.
    """

    def __init__(self, *streams):
        super().__init__()
        self.streams = streams
        self.current_stream = 0

    def read(self, n=-1):
        if n == 0:
            return b""

        if len(self.streams) == 0:
            return b""

        while self.current_stream < len(self.streams):
            stream = self.streams[self.current_stream]

            result = stream.read(n)
            if result is None or len(result) != 0:
                return result
            else:
                self.current_stream += 1

        return b""

    def readinto(self, b):
        n = len(b)
        read = self.read(n)
        b[: len(read)] = read
        return len(read)

    def close(self):
        for stream in self.streams:
            try:
                stream.close()
            except Exception:
                pass

    def readable(self, *args, **kwargs):
        return True

    def seekable(self, *args, **kwargs):
        return False

    def writable(self, *args, **kwargs):
        return False


class LineProcessorStream(io.RawIOBase):
    """
    While reading from this stream the provided `input_stream` is read line by line, calling the (overridable) method
    :meth:`.process_line` for each read line.

    Sub classes can thus modify the contents of the `input_stream` in line, while it is being read. Keep in mind that
    ``process_line`` will receive the line as a byte stream - if underlying code needs to operate on unicode you'll need
    to do the decoding yourself.

    Arguments:
        input_stream (io.RawIOBase): The stream to process on the fly.
    """

    def __init__(self, input_stream):
        super().__init__()
        self.input_stream = io.BufferedReader(input_stream)
        self.leftover = bytearray()

    def read(self, n=-1):
        if n == 0:
            return b""

        result = bytearray()
        while len(result) < n or n == -1:
            # add left over from previous loop
            bytes_left = (n - len(result)) if n != -1 else -1
            if bytes_left != -1 and bytes_left < len(self.leftover):
                # only s
                result += self.leftover[:bytes_left]
                self.leftover = self.leftover[bytes_left:]
                break
            else:
                result += self.leftover
                self.leftover = bytearray()

            # read one line from the underlying stream
            processed_line = None
            while processed_line is None:
                line = self.input_stream.readline()
                if not line:
                    break
                processed_line = self.process_line(line)
            if processed_line is None:
                break

            bytes_left = (n - len(result)) if n != -1 else -1
            if bytes_left != -1 and bytes_left < len(processed_line):
                result += processed_line[:bytes_left]
                self.leftover = processed_line[bytes_left:]
                break
            else:
                result += processed_line

        return bytes(result)

    def readinto(self, b):
        n = len(b)
        read = self.read(n)
        b[: len(read)] = read
        return len(read)

    def process_line(self, line):
        """
        Called from the `read` Method of this stream with each line read from `self.input_stream`.

        By returning ``None`` the line will not be returned from the read stream, effectively being stripped from the
        wrapper `input_stream`.

        Arguments:
            line (bytes): The line as read from `self.input_stream` in byte representation

        Returns:
            bytes or None: The processed version of the line (might also be multiple lines), or None if the line is to be
                stripped from the processed stream.
        """
        return line

    def close(self):
        self.input_stream.close()

    def readable(self, *args, **kwargs):
        return True

    def seekable(self, *args, **kwargs):
        return False

    def writable(self, *args, **kwargs):
        return False



# coding=utf8

CYRILLIC = {       # instead of:
    u'ё': u'e',    # io / yo
    u'у': u'y',    # u
    u'х': u'h',    # kh
    u'щ': u'sch',  # shch
    u'ю': u'u',    # iu / yu
    u'я': u'ya',   # ia
}

GERMAN = {         # instead of:
    u'ä': u'ae',   # a
    u'ö': u'oe',   # o
    u'ü': u'ue',   # u
}

GREEK = {          # instead of:
    u'Ξ': u'X',    # Ks
    u'χ': u'ch',   # kh

    u'ϒ': u'Y',    # U
    u'υ': u'y',    # u
    u'ύ': u'y',    # ...
    u'ϋ': u'y',
    u'ΰ': u'y',
}



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

import json
from collections import OrderedDict
from typing import Any, Callable, Dict

try:
    from typing import OrderedDict as OrderedDictType
except ImportError:
    OrderedDictType = Dict  # py3.7.{0,1}

from frozendict import frozendict

from octoprint.util import to_unicode


class JsonEncoding:
    encoders: OrderedDictType[type, Callable[[Any], Any]] = OrderedDict()

    @classmethod
    def add_encoder(cls, typ: type, encoder: Callable[[Any], Any]) -> None:
        """
        Add an encoder for a type.

        :param typ: the type to add an encoder for
        :param encoder: the encoder. Must take a single argument and return a
            tuple (name, parameters...)
        """
        cls.encoders[typ] = encoder

    @classmethod
    def remove_encoder(cls, typ):
        try:
            del cls.encoders[typ]
        except KeyError:
            pass

    @classmethod
    def dumps(cls, obj: Any) -> str:
        """
        Dump an object to JSON, handles additional types that the JSON encoder can't, like
        bytes and frozendicts.
        """
        return json.dumps(
            obj,
            default=cls.encode,
            separators=(",", ":"),
            indent=None,
            allow_nan=False,
        )

    @classmethod
    def loads(cls, s: str) -> Any:
        return json.loads(s)

    @classmethod
    def encode(cls, obj):
        for type, encoder in cls.encoders.items():
            if isinstance(obj, type):
                return encoder(obj)
        raise TypeError(f"Unserializable type {type(obj)}")


JsonEncoding.add_encoder(frozendict, lambda obj: dict(obj))
JsonEncoding.add_encoder(bytes, lambda obj: to_unicode(obj))

dumps = JsonEncoding.dumps
loads = JsonEncoding.loads



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

from enum import Enum
from typing import Dict, List, Optional

from octoprint.schema import BaseModel
from octoprint.vendor.with_attrs_docs import with_attrs_docs

CONST_15MIN = 15 * 60
CONST_1GB = 1024 * 1024 * 1024
CONST_500MB = 500 * 1024 * 1024
CONST_200MB = 200 * 1024 * 1024
CONST_100KB = 100 * 1024


@with_attrs_docs
class ReverseProxyConfig(BaseModel):
    prefixHeader: Optional[str] = None
    """The request header from which to determine the URL prefix under which OctoPrint is served by the reverse proxy."""

    schemeHeader: Optional[str] = None
    """The request header from which to determine the scheme (http or https) under which a specific request to OctoPrint was made to the reverse proxy."""

    hostHeader: Optional[str] = None
    """The request header from which to determine the host under which OctoPrint is served by the reverse proxy."""

    serverHeader: Optional[str] = None

    portHeader: Optional[str] = None

    prefixFallback: Optional[str] = None
    """Use this option to define an optional URL prefix (with a leading /, so absolute to your server's root) under which to run OctoPrint. This should only be needed if you want to run OctoPrint behind a reverse proxy under a different root endpoint than `/` and can't configure said reverse proxy to send a prefix HTTP header (X-Script-Name by default, see above) with forwarded requests."""

    schemeFallback: Optional[str] = None
    """Use this option to define an optional forced scheme (http or https) under which to run OctoPrint. This should only be needed if you want to run OctoPrint behind a reverse proxy that also does HTTPS determination but can't configure said reverse proxy to send a scheme HTTP header (X-Scheme by default, see above) with forwarded requests."""

    hostFallback: Optional[str] = None
    """Use this option to define an optional forced host under which to run OctoPrint. This should only be needed if you want to run OctoPrint behind a reverse proxy with a different hostname than OctoPrint itself but can't configure said reverse proxy to send a host HTTP header (X-Forwarded-Host by default, see above) with forwarded requests."""

    serverFallback: Optional[str] = None

    portFallback: Optional[str] = None

    trustedDownstream: List[str] = []
    """List of trusted downstream servers for which to ignore the IP address when trying to determine the connecting client's IP address. If you have OctoPrint behind more than one reverse proxy you should add their IPs here so that they won't be interpreted as the client's IP. One reverse proxy will be handled correctly by default."""


@with_attrs_docs
class UploadsConfig(BaseModel):
    maxSize: int = CONST_1GB
    """Maximum size of uploaded files in bytes, defaults to 1GB."""

    nameSuffix: str = "name"
    """Suffix used for storing the filename in the file upload headers when streaming uploads."""

    pathSuffix: str = "path"
    """Suffix used for storing the path to the temporary file in the file upload headers when streaming uploads."""


@with_attrs_docs
class CommandsConfig(BaseModel):
    systemShutdownCommand: Optional[str] = None
    """Command to shut down the system OctoPrint is running on."""

    systemRestartCommand: Optional[str] = None
    """Command to restart the system OctoPrint is running on."""

    serverRestartCommand: Optional[str] = None
    """Command to restart OctoPrint."""

    localPipCommand: Optional[str] = None
    """pip command associated with OctoPrint, used for installing plugins and updates, if unset (default) the command will be autodetected based on the current python executable - unless you have a really special setup this is the right way to do it and there should be no need to ever touch this setting."""


@with_attrs_docs
class OnlineCheckConfig(BaseModel):
    enabled: Optional[bool] = None
    """Whether the online check is enabled. Ships unset, the user will be asked to make a decision as part of the setup wizard."""

    interval: int = CONST_15MIN
    """Interval in which to check for online connectivity (in seconds), defaults to 15 minutes."""

    host: str = "1.1.1.1"
    """DNS host against which to check, defaults to Cloudflare's DNS."""

    port: int = 53
    """DNS port against which to check, defaults to the standard DNS port."""

    name: str = "octoprint.org"
    """Host name for which to check name resolution, defaults to OctoPrint's main domain."""


@with_attrs_docs
class PluginBlacklistConfig(BaseModel):
    enabled: Optional[bool] = None
    """Whether use of the blacklist is enabled. If unset, the user will be asked to make a decision as part of the setup wizard."""

    url: str = "https://plugins.octoprint.org/blacklist.json"
    """The URL from which to fetch the blacklist."""

    ttl: int = CONST_15MIN
    """Time to live of the cached blacklist, in seconds (default: 15 minutes)."""

    timeout: float = 3.05
    """Timeout for fetching the blacklist, in seconds (default: 3.05 seconds)."""


@with_attrs_docs
class DiskspaceConfig(BaseModel):
    warning: int = CONST_500MB
    """Threshold (bytes) after which to consider disk space becoming sparse, defaults to 500MB."""

    critical: int = CONST_200MB
    """Threshold (bytes) after which to consider disk space becoming critical, defaults to 200MB."""


@with_attrs_docs
class PreemptiveCacheConfig(BaseModel):
    exceptions: List[str] = []
    """Which server paths to exclude from the preemptive cache, e.g. `/some/path`."""

    until: int = 7
    """How many days to leave unused entries in the preemptive cache config."""


@with_attrs_docs
class IpCheckConfig(BaseModel):
    enabled: bool = True
    """Whether to enable the check."""

    trustedSubnets: List[str] = []
    """Additional non-local subnets to consider trusted, in CIDR notation, e.g. `192.168.1.0/24`."""


class SameSiteEnum(str, Enum):
    strict = "Strict"
    lax = "Lax"
    none = "None"


@with_attrs_docs
class CookiesConfig(BaseModel):
    secure: bool = False
    """Whether to set the `Secure` flag to true on cookies. Only set to true if you are running OctoPrint behind a reverse proxy taking care of SSL termination."""

    samesite: Optional[SameSiteEnum] = SameSiteEnum.lax
    """`SameSite` setting to use on the cookies. Possible values are `None`, `Lax` and `Strict`. Defaults to `Lax`. Be advised that if forced unset, this has security implications as many browsers now default to `Lax` unless you configure cookies to be set with `Secure` flag set, explicitly set `SameSite` setting here and also serve OctoPrint over https. The `Lax` setting is known to cause with embedding OctoPrint in frames. See also ["Feature: Cookies default to SameSite=Lax"](https://www.chromestatus.com/feature/5088147346030592), ["Feature: Reject insecure SameSite=None cookies"](https://www.chromestatus.com/feature/5633521622188032) and [issue #3482](https://github.com/OctoPrint/OctoPrint/issues/3482)."""


@with_attrs_docs
class ServerConfig(BaseModel):
    host: Optional[str] = None
    """Use this option to define the host to which to bind the server. If unset, OctoPrint will attempt to bind on all available interfaces, IPv4 and v6 unless either is disabled."""

    port: int = 5000
    """Use this option to define the port to which to bind the server."""

    firstRun: bool = True
    """If this option is true, OctoPrint will show the First Run wizard and set the setting to false after that completes."""

    startOnceInSafeMode: bool = False
    """If this option is true, OctoPrint will enable safe mode on the next server start and reset the setting to false"""

    ignoreIncompleteStartup: bool = False
    """Set this to true to make OctoPrint ignore incomplete startups. Helpful for development."""

    incompleteStartup: bool = False
    """Signals to OctoPrint that the last startup was incomplete. OctoPrint will then startup in safe mode."""

    seenWizards: Dict[str, str] = {}

    secretKey: Optional[str] = None
    """Secret key for encrypting cookies and such, randomly generated on first run."""

    heartbeat: int = CONST_15MIN

    reverseProxy: ReverseProxyConfig = ReverseProxyConfig()
    """Settings if OctoPrint is running behind a reverse proxy (haproxy, nginx, apache, ...) that doesn't correctly set the [required headers](https://community.octoprint.org/t/reverse-proxy-configuration-examples/1107). These are necessary in order to make OctoPrint generate correct external URLs so that AJAX requests and download URLs work, and so that client IPs are read correctly."""

    uploads: UploadsConfig = UploadsConfig()
    """Settings for file uploads to OctoPrint, such as maximum allowed file size and header suffixes to use for streaming uploads. OctoPrint does some nifty things internally in order to allow streaming of large file uploads to the application rather than just storing them in memory. For that it needs to do some rewriting of the incoming upload HTTP requests, storing the uploaded file to a temporary location on disk and then sending an internal request to the application containing the original filename and the location of the temporary file."""

    maxSize: int = CONST_100KB
    """Maximum size of requests other than file uploads in bytes, defaults to 100KB."""

    commands: CommandsConfig = CommandsConfig()
    """Commands to restart/shutdown octoprint or the system it's running on."""

    onlineCheck: OnlineCheckConfig = OnlineCheckConfig()
    """Configuration of the regular online connectivity check."""

    pluginBlacklist: PluginBlacklistConfig = PluginBlacklistConfig()
    """Configuration of the plugin blacklist."""

    diskspace: DiskspaceConfig = DiskspaceConfig()
    """Settings of when to display what disk space warning."""

    preemptiveCache: PreemptiveCacheConfig = PreemptiveCacheConfig()
    """Configuration of the preemptive cache."""

    ipCheck: IpCheckConfig = IpCheckConfig()
    """Configuration of the client IP check to warn about connections from external networks."""

    allowFraming: bool = False
    """Whether to allow OctoPrint to be embedded in a frame or not. Note that depending on your setup you might have to set SameSite to None, Secure to true and serve OctoPrint through a reverse proxy that enables https for cookies and thus logging in to work."""

    cookies: CookiesConfig = CookiesConfig()
    """Settings for further configuration of the cookies that OctoPrint sets (login, remember me, ...)."""

    allowedLoginRedirectPaths: List[str] = []
    """List of paths that are allowed to be used as redirect targets for the login page, in addition to the default ones (`/`, `/recovery/` and `/plugin/appkeys/auth/`)"""



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

from enum import Enum

from octoprint.schema import BaseModel
from octoprint.vendor.with_attrs_docs import with_attrs_docs


class StylesheetEnum(str, Enum):
    css = "css"
    less = "less"


@with_attrs_docs
class DevelWebassetsConfig(BaseModel):
    bundle: bool = True
    """If set to true, OctoPrint will merge all JS, all CSS and all Less files into one file per type to reduce request count. Setting it to false will load all assets individually. Note: if this is set to false, no minification will take place regardless of the `minify` setting."""

    clean_on_startup: bool = True
    """Whether to delete generated web assets on server startup (forcing a regeneration)."""

    minify: bool = True
    """If set to true, OctoPrint will the core and library javascript assets. Note: if `bundle` is set to false, no minification will take place either."""

    minify_plugins: bool = False
    """If set to true, OctoPrint will also minify the third party plugin javascript assets. Note: if `bundle` or `minify` are set to false, no minification of the plugin assets will take place either."""


class DevelCacheConfig(BaseModel):
    enabled: bool = True
    """Whether to enable caching. Defaults to true. Setting it to false will cause the UI to always be fully rerendered on request to `/` on the server."""

    preemptive: bool = True
    """Whether to enable the preemptive cache."""


@with_attrs_docs
class DevelConfig(BaseModel):
    stylesheet: StylesheetEnum = StylesheetEnum.css
    """Settings for stylesheet preference. OctoPrint will prefer to use the stylesheet type specified here. Usually (on a production install) that will be the compiled css (default). Developers may specify less here too."""

    cache: DevelCacheConfig = DevelCacheConfig()
    """Settings for OctoPrint's internal caching."""

    webassets: DevelWebassetsConfig = DevelWebassetsConfig()
    """Settings for OctoPrint's web asset merging and minifying."""

    useFrozenDictForPrinterState: bool = True

    showLoadingAnimation: bool = True
    """Enable or disable the loading animation."""

    sockJsConnectTimeout: float = 30
    pluginTimings: bool = False

    enableRateLimiter: bool = True
    """Enable or disable the rate limiter. Careful, disabling this reduces security."""

    enableCsrfProtection: bool = True
    """Enable or disable the CSRF protection. Careful, disabling this reduces security."""



__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import hashlib
import logging
import os
import threading
from urllib.parse import quote as urlquote

import psutil
from flask import abort, jsonify, make_response, request, url_for

import octoprint.filemanager
import octoprint.filemanager.storage
import octoprint.filemanager.util
import octoprint.slicing
from octoprint.access.permissions import Permissions
from octoprint.events import Events
from octoprint.filemanager.destinations import FileDestinations
from octoprint.filemanager.storage import StorageError
from octoprint.server import (
    NO_CONTENT,
    current_user,
    eventManager,
    fileManager,
    printer,
    slicingManager,
)
from octoprint.server.api import api
from octoprint.server.util.flask import (
    get_json_command_from_request,
    no_firstrun_access,
    with_revalidation_checking,
)
from octoprint.settings import settings, valid_boolean_trues
from octoprint.util import sv, time_this

# ~~ GCODE file handling

_file_cache = {}
_file_cache_mutex = threading.RLock()

_DATA_FORMAT_VERSION = "v2"


def _clear_file_cache():
    with _file_cache_mutex:
        _file_cache.clear()


def _create_lastmodified(path, recursive):
    path = path[len("/api/files") :]
    if path.startswith("/"):
        path = path[1:]

    if path == "":
        # all storages involved
        lms = [0]
        for storage in fileManager.registered_storages:
            try:
                lms.append(fileManager.last_modified(storage, recursive=recursive))
            except Exception:
                logging.getLogger(__name__).exception(
                    "There was an error retrieving the last modified data from storage {}".format(
                        storage
                    )
                )
                lms.append(None)

        if any(filter(lambda x: x is None, lms)):
            # we return None if ANY of the involved storages returned None
            return None

        # if we reach this point, we return the maximum of all dates
        return max(lms)

    else:
        if "/" in path:
            storage, path_in_storage = path.split("/", 1)
        else:
            storage = path
            path_in_storage = None

        try:
            return fileManager.last_modified(
                storage, path=path_in_storage, recursive=recursive
            )
        except Exception:
            logging.getLogger(__name__).exception(
                "There was an error retrieving the last modified data from storage {} and path {}".format(
                    storage, path_in_storage
                )
            )
            return None


def _create_etag(path, filter, recursive, lm=None):
    if lm is None:
        lm = _create_lastmodified(path, recursive)

    if lm is None:
        return None

    hash = hashlib.sha1()

    def hash_update(value):
        value = value.encode("utf-8")
        hash.update(value)

    hash_update(str(lm))
    hash_update(str(filter))
    hash_update(str(recursive))

    path = path[len("/api/files") :]
    if path.startswith("/"):
        path = path[1:]

    if "/" in path:
        storage, _ = path.split("/", 1)
    else:
        storage = path

    if path == "" or storage == FileDestinations.SDCARD:
        # include sd data in etag
        hash_update(repr(sorted(printer.get_sd_files(), key=lambda x: sv(x["name"]))))

    hash_update(_DATA_FORMAT_VERSION)  # increment version if we change the API format

    return hash.hexdigest()


@api.route("/files", methods=["GET"])
@Permissions.FILES_LIST.require(403)
@with_revalidation_checking(
    etag_factory=lambda lm=None: _create_etag(
        request.path,
        request.values.get("filter", False),
        request.values.get("recursive", False),
        lm=lm,
    ),
    lastmodified_factory=lambda: _create_lastmodified(
        request.path, request.values.get("recursive", False)
    ),
    unless=lambda: request.values.get("force", False)
    or request.values.get("_refresh", False),
)
def readGcodeFiles():
    filter = request.values.get("filter", False)
    recursive = request.values.get("recursive", "false") in valid_boolean_trues
    force = request.values.get("force", "false") in valid_boolean_trues

    files = _getFileList(
        FileDestinations.LOCAL,
        filter=filter,
        recursive=recursive,
        allow_from_cache=not force,
    )
    files.extend(_getFileList(FileDestinations.SDCARD, allow_from_cache=not force))

    usage = psutil.disk_usage(settings().getBaseFolder("uploads", check_writable=False))
    return jsonify(files=files, free=usage.free, total=usage.total)


@api.route("/files/test", methods=["POST"])
@Permissions.FILES_LIST.require(403)
def runFilesTest():
    valid_commands = {
        "sanitize": ["storage", "path", "filename"],
        "exists": ["storage", "path", "filename"],
    }

    command, data, response = get_json_command_from_request(request, valid_commands)
    if response is not None:
        return response

    def sanitize(storage, path, filename):
        sanitized_path = fileManager.sanitize_path(storage, path)
        sanitized_name = fileManager.sanitize_name(storage, filename)
        joined = fileManager.join_path(storage, sanitized_path, sanitized_name)
        return sanitized_path, sanitized_name, joined

    if command == "sanitize":
        _, _, sanitized = sanitize(data["storage"], data["path"], data["filename"])
        return jsonify(sanitized=sanitized)
    elif command == "exists":
        storage = data["storage"]
        path = data["path"]
        filename = data["filename"]

        sanitized_path, _, sanitized = sanitize(storage, path, filename)

        exists = fileManager.file_exists(storage, sanitized)
        if exists:
            suggestion = filename
            name, ext = os.path.splitext(filename)
            counter = 0
            while fileManager.file_exists(
                storage,
                fileManager.join_path(
                    storage,
                    sanitized_path,
                    fileManager.sanitize_name(storage, suggestion),
                ),
            ):
                counter += 1
                suggestion = f"{name}_{counter}{ext}"
            return jsonify(exists=True, suggestion=suggestion)
        else:
            return jsonify(exists=False)


@api.route("/files/<string:origin>", methods=["GET"])
@Permissions.FILES_LIST.require(403)
@with_revalidation_checking(
    etag_factory=lambda lm=None: _create_etag(
        request.path,
        request.values.get("filter", False),
        request.values.get("recursive", False),
        lm=lm,
    ),
    lastmodified_factory=lambda: _create_lastmodified(
        request.path, request.values.get("recursive", False)
    ),
    unless=lambda: request.values.get("force", False)
    or request.values.get("_refresh", False),
)
def readGcodeFilesForOrigin(origin):
    if origin not in [FileDestinations.LOCAL, FileDestinations.SDCARD]:
        abort(404)

    filter = request.values.get("filter", False)
    recursive = request.values.get("recursive", "false") in valid_boolean_trues
    force = request.values.get("force", "false") in valid_boolean_trues

    files = _getFileList(
        origin, filter=filter, recursive=recursive, allow_from_cache=not force
    )

    if origin == FileDestinations.LOCAL:
        usage = psutil.disk_usage(
            settings().getBaseFolder("uploads", check_writable=False)
        )
        return jsonify(files=files, free=usage.free, total=usage.total)
    else:
        return jsonify(files=files)


@api.route("/files/<string:target>/<path:filename>", methods=["GET"])
@Permissions.FILES_LIST.require(403)
@with_revalidation_checking(
    etag_factory=lambda lm=None: _create_etag(
        request.path,
        request.values.get("filter", False),
        request.values.get("recursive", False),
        lm=lm,
    ),
    lastmodified_factory=lambda: _create_lastmodified(
        request.path, request.values.get("recursive", False)
    ),
    unless=lambda: request.values.get("force", False)
    or request.values.get("_refresh", False),
)
def readGcodeFile(target, filename):
    if target not in [FileDestinations.LOCAL, FileDestinations.SDCARD]:
        abort(404)

    if not _validate(target, filename):
        abort(404)

    recursive = False
    if "recursive" in request.values:
        recursive = request.values["recursive"] in valid_boolean_trues

    file = _getFileDetails(target, filename, recursive=recursive)
    if not file:
        abort(404)

    return jsonify(file)


def _getFileDetails(origin, path, recursive=True):
    parent, path = os.path.split(path)
    files = _getFileList(origin, path=parent, recursive=recursive, level=1)

    for f in files:
        if f["name"] == path:
            return f
    else:
        return None


@time_this(
    logtarget=__name__ + ".timings",
    message="{func}({func_args},{func_kwargs}) took {timing:.2f}ms",
    incl_func_args=True,
    log_enter=True,
    message_enter="Entering {func}({func_args},{func_kwargs})...",
)
def _getFileList(
    origin, path=None, filter=None, recursive=False, level=0, allow_from_cache=True
):
    if origin == FileDestinations.SDCARD:
        sdFileList = printer.get_sd_files(refresh=not allow_from_cache)

        files = []
        if sdFileList is not None:
            for f in sdFileList:
                type_path = octoprint.filemanager.get_file_type(f["name"])
                if not type_path:
                    # only supported extensions
                    continue
                else:
                    file_type = type_path[0]

                file = {
                    "type": file_type,
                    "typePath": type_path,
                    "name": f["name"],
                    "display": f["display"] if f["display"] else f["name"],
                    "path": f["name"],
                    "origin": FileDestinations.SDCARD,
                    "refs": {
                        "resource": url_for(
                            ".readGcodeFile",
                            target=FileDestinations.SDCARD,
                            filename=f["name"],
                            _external=True,
                        )
                    },
                }
                if f["size"] is not None:
                    file.update({"size": f["size"]})
                if f["date"] is not None:
                    file.update({"date": f["date"]})
                files.append(file)
    else:
        # PERF: Only retrieve the extension tree once
        extension_tree = octoprint.filemanager.full_extension_tree()

        filter_func = None
        if filter:
            filter_func = lambda entry, entry_data: octoprint.filemanager.valid_file_type(
                entry, type=filter, tree=extension_tree
            )

        with _file_cache_mutex:
            cache_key = f"{origin}:{path}:{recursive}:{filter}"
            files, lastmodified = _file_cache.get(cache_key, ([], None))
            # recursive needs to be True for lastmodified queries so we get lastmodified of whole subtree - #3422
            if (
                not allow_from_cache
                or lastmodified is None
                or lastmodified
                < fileManager.last_modified(origin, path=path, recursive=True)
            ):
                files = list(
                    fileManager.list_files(
                        origin,
                        path=path,
                        filter=filter_func,
                        recursive=recursive,
                        level=level,
                        force_refresh=not allow_from_cache,
                    )[origin].values()
                )
                lastmodified = fileManager.last_modified(
                    origin, path=path, recursive=True
                )
                _file_cache[cache_key] = (files, lastmodified)

        def analyse_recursively(files, path=None):
            if path is None:
                path = ""

            result = []
            for file_or_folder in files:
                # make a shallow copy in order to not accidentally modify the cached data
                file_or_folder = dict(file_or_folder)

                file_or_folder["origin"] = FileDestinations.LOCAL

                if file_or_folder["type"] == "folder":
                    if "children" in file_or_folder:
                        file_or_folder["children"] = analyse_recursively(
                            file_or_folder["children"].values(),
                            path + file_or_folder["name"] + "/",
                        )

                    file_or_folder["refs"] = {
                        "resource": url_for(
                            ".readGcodeFile",
                            target=FileDestinations.LOCAL,
                            filename=path + file_or_folder["name"],
                            _external=True,
                        )
                    }
                else:
                    if (
                        "analysis" in file_or_folder
                        and octoprint.filemanager.valid_file_type(
                            file_or_folder["name"], type="gcode", tree=extension_tree
                        )
                    ):
                        file_or_folder["gcodeAnalysis"] = file_or_folder["analysis"]
                        del file_or_folder["analysis"]

                    if (
                        "history" in file_or_folder
                        and octoprint.filemanager.valid_file_type(
                            file_or_folder["name"], type="gcode", tree=extension_tree
                        )
                    ):
                        # convert print log
                        history = file_or_folder["history"]
                        del file_or_folder["history"]
                        success = 0
                        failure = 0
                        last = None
                        for entry in history:
                            success += 1 if "success" in entry and entry["success"] else 0
                            failure += (
                                1 if "success" in entry and not entry["success"] else 0
                            )
                            if not last or (
                                "timestamp" in entry
                                and "timestamp" in last
                                and entry["timestamp"] > last["timestamp"]
                            ):
                                last = entry
                        if last:
                            prints = {
                                "success": success,
                                "failure": failure,
                                "last": {
                                    "success": last["success"],
                                    "date": last["timestamp"],
                                },
                            }
                            if "printTime" in last:
                                prints["last"]["printTime"] = last["printTime"]
                            file_or_folder["prints"] = prints

                    file_or_folder["refs"] = {
                        "resource": url_for(
                            ".readGcodeFile",
                            target=FileDestinations.LOCAL,
                            filename=file_or_folder["path"],
                            _external=True,
                        ),
                        "download": url_for("index", _external=True)
                        + "downloads/files/"
                        + FileDestinations.LOCAL
                        + "/"
                        + urlquote(file_or_folder["path"]),
                    }

                result.append(file_or_folder)

            return result

        files = analyse_recursively(files)

    return files


def _verifyFileExists(origin, filename):
    if origin == FileDestinations.SDCARD:
        return filename in (x["name"] for x in printer.get_sd_files())
    else:
        return fileManager.file_exists(origin, filename)


def _verifyFolderExists(origin, foldername):
    if origin == FileDestinations.SDCARD:
        return False
    else:
        return fileManager.folder_exists(origin, foldername)


def _isBusy(target, path):
    currentOrigin, currentPath = _getCurrentFile()
    if (
        currentPath is not None
        and currentOrigin == target
        and fileManager.file_in_path(FileDestinations.LOCAL, path, currentPath)
        and (printer.is_printing() or printer.is_paused())
    ):
        return True

    return any(
        target == x[0] and fileManager.file_in_path(FileDestinations.LOCAL, path, x[1])
        for x in fileManager.get_busy_files()
    )


@api.route("/files/<string:target>", methods=["POST"])
@no_firstrun_access
@Permissions.FILES_UPLOAD.require(403)
def uploadGcodeFile(target):
    input_name = "file"
    input_upload_name = (
        input_name + "." + settings().get(["server", "uploads", "nameSuffix"])
    )
    input_upload_path = (
        input_name + "." + settings().get(["server", "uploads", "pathSuffix"])
    )
    if input_upload_name in request.values and input_upload_path in request.values:
        if target not in [FileDestinations.LOCAL, FileDestinations.SDCARD]:
            abort(404)

        upload = octoprint.filemanager.util.DiskFileWrapper(
            request.values[input_upload_name], request.values[input_upload_path]
        )

        # Store any additional user data the caller may have passed.
        userdata = None
        if "userdata" in request.values:
            import json

            try:
                userdata = json.loads(request.values["userdata"])
            except Exception:
                abort(400, description="userdata contains invalid JSON")

        # check preconditions for SD upload
        if target == FileDestinations.SDCARD and not settings().getBoolean(
            ["feature", "sdSupport"]
        ):
            abort(404)

        sd = target == FileDestinations.SDCARD
        if sd:
            # validate that all preconditions for SD upload are met before attempting it
            if not (
                printer.is_operational()
                and not (printer.is_printing() or printer.is_paused())
            ):
                abort(
                    409,
                    description="Can not upload to SD card, printer is either not operational or already busy",
                )
            if not printer.is_sd_ready():
                abort(409, description="Can not upload to SD card, not yet initialized")

        # evaluate select and print parameter and if set check permissions & preconditions
        # and adjust as necessary
        #
        # we do NOT abort(409) here since this would be a backwards incompatible behaviour change
        # on the API, but instead return the actually effective select and print flags in the response
        #
        # note that this behaviour might change in a future API version
        select_request = (
            "select" in request.values
            and request.values["select"] in valid_boolean_trues
            and Permissions.FILES_SELECT.can()
        )
        print_request = (
            "print" in request.values
            and request.values["print"] in valid_boolean_trues
            and Permissions.PRINT.can()
        )

        to_select = select_request
        to_print = print_request
        if (to_select or to_print) and not (
            printer.is_operational()
            and not (printer.is_printing() or printer.is_paused())
        ):
            # can't select or print files if not operational or ready
            to_select = to_print = False

        # determine future filename of file to be uploaded, abort if it can't be uploaded
        try:
            # FileDestinations.LOCAL = should normally be target, but can't because SDCard handling isn't implemented yet
            canonPath, canonFilename = fileManager.canonicalize(
                FileDestinations.LOCAL, upload.filename
            )
            if request.values.get("path"):
                canonPath = request.values.get("path")
            if request.values.get("filename"):
                canonFilename = request.values.get("filename")

            futurePath = fileManager.sanitize_path(FileDestinations.LOCAL, canonPath)
            futureFilename = fileManager.sanitize_name(
                FileDestinations.LOCAL, canonFilename
            )
        except Exception:
            canonFilename = None
            futurePath = None
            futureFilename = None

        if futureFilename is None:
            abort(400, description="Can not upload file, invalid file name")

        # prohibit overwriting currently selected file while it's being printed
        futureFullPath = fileManager.join_path(
            FileDestinations.LOCAL, futurePath, futureFilename
        )
        futureFullPathInStorage = fileManager.path_in_storage(
            FileDestinations.LOCAL, futureFullPath
        )

        if not printer.can_modify_file(futureFullPathInStorage, sd):
            abort(
                409,
                description="Trying to overwrite file that is currently being printed",
            )

        if (
            fileManager.file_exists(FileDestinations.LOCAL, futureFullPathInStorage)
            and request.values.get("noOverwrite") in valid_boolean_trues
        ):
            abort(409, description="File already exists and noOverwrite was set")

        if (
            fileManager.file_exists(FileDestinations.LOCAL, futureFullPathInStorage)
            and not Permissions.FILES_DELETE.can()
        ):
            abort(
                403,
                description="File already exists, cannot overwrite due to a lack of permissions",
            )

        reselect = printer.is_current_file(futureFullPathInStorage, sd)

        user = current_user.get_name()

        def fileProcessingFinished(filename, absFilename, destination):
            """
            Callback for when the file processing (upload, optional slicing, addition to analysis queue) has
            finished.

            Depending on the file's destination triggers either streaming to SD card or directly calls to_select.
            """

            if (
                destination == FileDestinations.SDCARD
                and octoprint.filemanager.valid_file_type(filename, "machinecode")
            ):
                return filename, printer.add_sd_file(
                    filename,
                    absFilename,
                    on_success=selectAndOrPrint,
                    tags={"source:api", "api:files.sd"},
                )
            else:
                selectAndOrPrint(filename, absFilename, destination)
                return filename

        def selectAndOrPrint(filename, absFilename, destination):
            """
            Callback for when the file is ready to be selected and optionally printed. For SD file uploads this is only
            the case after they have finished streaming to the printer, which is why this callback is also used
            for the corresponding call to addSdFile.

            Selects the just uploaded file if either to_select or to_print are True, or if the
            exact file is already selected, such reloading it.
            """
            if octoprint.filemanager.valid_file_type(added_file, "gcode") and (
                to_select or to_print or reselect
            ):
                printer.select_file(
                    absFilename,
                    destination == FileDestinations.SDCARD,
                    to_print,
                    user,
                )

        try:
            added_file = fileManager.add_file(
                FileDestinations.LOCAL,
                futureFullPathInStorage,
                upload,
                allow_overwrite=True,
                display=canonFilename,
            )
        except (OSError, StorageError) as e:
            _abortWithException(e)
        else:
            filename = fileProcessingFinished(
                added_file,
                fileManager.path_on_disk(FileDestinations.LOCAL, added_file),
                target,
            )
            done = not sd

        if userdata is not None:
            # upload included userdata, add this now to the metadata
            fileManager.set_additional_metadata(
                FileDestinations.LOCAL, added_file, "userdata", userdata
            )

        sdFilename = None
        if isinstance(filename, tuple):
            filename, sdFilename = filename

        payload = {
            "name": futureFilename,
            "path": filename,
            "target": target,
            "select": select_request,
            "print": print_request,
            "effective_select": to_select,
            "effective_print": to_print,
        }
        if userdata is not None:
            payload["userdata"] = userdata
        eventManager.fire(Events.UPLOAD, payload)

        files = {}
        location = url_for(
            ".readGcodeFile",
            target=FileDestinations.LOCAL,
            filename=filename,
            _external=True,
        )
        files.update(
            {
                FileDestinations.LOCAL: {
                    "name": futureFilename,
                    "path": filename,
                    "origin": FileDestinations.LOCAL,
                    "refs": {
                        "resource": location,
                        "download": url_for("index", _external=True)
                        + "downloads/files/"
                        + FileDestinations.LOCAL
                        + "/"
                        + urlquote(filename),
                    },
                }
            }
        )

        if sd and sdFilename:
            location = url_for(
                ".readGcodeFile",
                target=FileDestinations.SDCARD,
                filename=sdFilename,
                _external=True,
            )
            files.update(
                {
                    FileDestinations.SDCARD: {
                        "name": sdFilename,
                        "path": sdFilename,
                        "origin": FileDestinations.SDCARD,
                        "refs": {"resource": location},
                    }
                }
            )

        r = make_response(
            jsonify(
                files=files,
                done=done,
                effectiveSelect=to_select,
                effectivePrint=to_print,
            ),
            201,
        )
        r.headers["Location"] = location
        return r

    elif "foldername" in request.values:
        foldername = request.values["foldername"]

        if target not in [FileDestinations.LOCAL]:
            abort(400, description="target is invalid")

        canonPath, canonName = fileManager.canonicalize(target, foldername)
        futurePath = fileManager.sanitize_path(target, canonPath)
        futureName = fileManager.sanitize_name(target, canonName)
        if not futureName or not futurePath:
            abort(400, description="folder name is empty")

        if "path" in request.values and request.values["path"]:
            futurePath = fileManager.sanitize_path(
                FileDestinations.LOCAL, request.values["path"]
            )

        futureFullPath = fileManager.join_path(target, futurePath, futureName)
        if octoprint.filemanager.valid_file_type(futureName):
            abort(409, description="Can't create folder, please try another name")

        try:
            added_folder = fileManager.add_folder(
                target, futureFullPath, display=canonName
            )
        except (OSError, StorageError) as e:
            _abortWithException(e)

        location = url_for(
            ".readGcodeFile",
            target=FileDestinations.LOCAL,
            filename=added_folder,
            _external=True,
        )
        folder = {
            "name": futureName,
            "path": added_folder,
            "origin": target,
            "refs": {"resource": location},
        }

        r = make_response(jsonify(folder=folder, done=True), 201)
        r.headers["Location"] = location
        return r

    else:
        abort(400, description="No file to upload and no folder to create")


@api.route("/files/<string:target>/<path:filename>", methods=["POST"])
@no_firstrun_access
def gcodeFileCommand(filename, target):
    if target not in [FileDestinations.LOCAL, FileDestinations.SDCARD]:
        abort(404)

    if not _validate(target, filename):
        abort(404)

    # valid file commands, dict mapping command name to mandatory parameters
    valid_commands = {
        "select": [],
        "unselect": [],
        "slice": [],
        "analyse": [],
        "copy": ["destination"],
        "move": ["destination"],
    }

    command, data, response = get_json_command_from_request(request, valid_commands)
    if response is not None:
        return response

    user = current_user.get_name()

    if command == "select":
        with Permissions.FILES_SELECT.require(403):
            if not _verifyFileExists(target, filename):
                abort(404)

            # selects/loads a file
            if not octoprint.filemanager.valid_file_type(filename, type="machinecode"):
                abort(
                    415,
                    description="Cannot select file for printing, not a machinecode file",
                )

            if not printer.is_ready():
                abort(
                    409,
                    description="Printer is already printing, cannot select a new file",
                )

            printAfterLoading = False
            if "print" in data and data["print"] in valid_boolean_trues:
                with Permissions.PRINT.require(403):
                    if not printer.is_operational():
                        abort(
                            409,
                            description="Printer is not operational, cannot directly start printing",
                        )
                    printAfterLoading = True

            sd = False
            if target == FileDestinations.SDCARD:
                filenameToSelect = filename
                sd = True
            else:
                filenameToSelect = fileManager.path_on_disk(target, filename)
            printer.select_file(filenameToSelect, sd, printAfterLoading, user)

    elif command == "unselect":
        with Permissions.FILES_SELECT.require(403):
            if not printer.is_ready():
                return make_response(
                    "Printer is already printing, cannot unselect current file", 409
                )

            _, currentFilename = _getCurrentFile()
            if currentFilename is None:
                return make_response(
                    "Cannot unselect current file when there is no file selected", 409
                )

            if filename != currentFilename and filename != "current":
                return make_response(
                    "Only the currently selected file can be unselected", 400
                )

            printer.unselect_file()

    elif command == "slice":
        with Permissions.SLICE.require(403):
            if not _verifyFileExists(target, filename):
                abort(404)

            try:
                if "slicer" in data:
                    slicer = data["slicer"]
                    del data["slicer"]
                    slicer_instance = slicingManager.get_slicer(slicer)

                elif "cura" in slicingManager.registered_slicers:
                    slicer = "cura"
                    slicer_instance = slicingManager.get_slicer("cura")

                else:
                    abort(415, description="Cannot slice file, no slicer available")
            except octoprint.slicing.UnknownSlicer:
                abort(404)

            if not any(
                [
                    octoprint.filemanager.valid_file_type(filename, type=source_file_type)
                    for source_file_type in slicer_instance.get_slicer_properties().get(
                        "source_file_types", ["model"]
                    )
                ]
            ):
                abort(415, description="Cannot slice file, not a model file")

            cores = psutil.cpu_count()
            if (
                slicer_instance.get_slicer_properties().get("same_device", True)
                and (printer.is_printing() or printer.is_paused())
                and (cores is None or cores < 2)
            ):
                # slicer runs on same device as OctoPrint, slicing while printing is hence disabled
                abort(
                    409,
                    description="Cannot slice on this slicer while printing on single core systems or systems of unknown core count due to performance reasons",
                )

            if "destination" in data and data["destination"]:
                destination = data["destination"]
                del data["destination"]
            elif "gcode" in data and data["gcode"]:
                destination = data["gcode"]
                del data["gcode"]
            else:
                import os

                name, _ = os.path.splitext(filename)
                destination = (
                    name
                    + "."
                    + slicer_instance.get_slicer_properties().get(
                        "destination_extensions", ["gco", "gcode", "g"]
                    )[0]
                )

            full_path = destination
            if "path" in data and data["path"]:
                full_path = fileManager.join_path(target, data["path"], destination)
            else:
                path, _ = fileManager.split_path(target, filename)
                if path:
                    full_path = fileManager.join_path(target, path, destination)

            canon_path, canon_name = fileManager.canonicalize(target, full_path)
            sanitized_name = fileManager.sanitize_name(target, canon_name)

            if canon_path:
                full_path = fileManager.join_path(target, canon_path, sanitized_name)
            else:
                full_path = sanitized_name

            # prohibit overwriting the file that is currently being printed
            currentOrigin, currentFilename = _getCurrentFile()
            if (
                currentFilename == full_path
                and currentOrigin == target
                and (printer.is_printing() or printer.is_paused())
            ):
                abort(
                    409,
                    description="Trying to slice into file that is currently being printed",
                )

            if "profile" in data and data["profile"]:
                profile = data["profile"]
                del data["profile"]
            else:
                profile = None

            if "printerProfile" in data and data["printerProfile"]:
                printerProfile = data["printerProfile"]
                del data["printerProfile"]
            else:
                printerProfile = None

            if (
                "position" in data
                and data["position"]
                and isinstance(data["position"], dict)
                and "x" in data["position"]
                and "y" in data["position"]
            ):
                position = data["position"]
                del data["position"]
            else:
                position = None

            select_after_slicing = False
            if "select" in data and data["select"] in valid_boolean_trues:
                if not printer.is_operational():
                    abort(
                        409,
                        description="Printer is not operational, cannot directly select for printing",
                    )
                select_after_slicing = True

            print_after_slicing = False
            if "print" in data and data["print"] in valid_boolean_trues:
                if not printer.is_operational():
                    abort(
                        409,
                        description="Printer is not operational, cannot directly start printing",
                    )
                select_after_slicing = print_after_slicing = True

            override_keys = [
                k for k in data if k.startswith("profile.") and data[k] is not None
            ]
            overrides = {}
            for key in override_keys:
                overrides[key[len("profile.") :]] = data[key]

            def slicing_done(target, path, select_after_slicing, print_after_slicing):
                if select_after_slicing or print_after_slicing:
                    sd = False
                    if target == FileDestinations.SDCARD:
                        filenameToSelect = path
                        sd = True
                    else:
                        filenameToSelect = fileManager.path_on_disk(target, path)
                    printer.select_file(filenameToSelect, sd, print_after_slicing, user)

            try:
                fileManager.slice(
                    slicer,
                    target,
                    filename,
                    target,
                    full_path,
                    profile=profile,
                    printer_profile_id=printerProfile,
                    position=position,
                    overrides=overrides,
                    display=canon_name,
                    callback=slicing_done,
                    callback_args=(
                        target,
                        full_path,
                        select_after_slicing,
                        print_after_slicing,
                    ),
                )
            except octoprint.slicing.UnknownProfile:
                abort(404, description="Unknown profile")

            location = url_for(
                ".readGcodeFile",
                target=target,
                filename=full_path,
                _external=True,
            )
            result = {
                "name": destination,
                "path": full_path,
                "display": canon_name,
                "origin": FileDestinations.LOCAL,
                "refs": {
                    "resource": location,
                    "download": url_for("index", _external=True)
                    + "downloads/files/"
                    + target
                    + "/"
                    + urlquote(full_path),
                },
            }

            r = make_response(jsonify(result), 202)
            r.headers["Location"] = location
            return r

    elif command == "analyse":
        with Permissions.FILES_UPLOAD.require(403):
            if not _verifyFileExists(target, filename):
                abort(404)

            printer_profile = None
            if "printerProfile" in data and data["printerProfile"]:
                printer_profile = data["printerProfile"]

            if not fileManager.analyse(
                target, filename, printer_profile_id=printer_profile
            ):
                abort(400, description="No analysis possible")

    elif command == "copy" or command == "move":
        with Permissions.FILES_UPLOAD.require(403):
            # Copy and move are only possible on local storage
            if target not in [FileDestinations.LOCAL]:
                abort(400, description=f"Unsupported target for {command}")

            if not _verifyFileExists(target, filename) and not _verifyFolderExists(
                target, filename
            ):
                abort(404)

            path, name = fileManager.split_path(target, filename)

            destination = data["destination"]
            dst_path, dst_name = fileManager.split_path(target, destination)
            sanitized_destination = fileManager.join_path(
                target, dst_path, fileManager.sanitize_name(target, dst_name)
            )

            # Check for exception thrown by _verifyFolderExists, if outside the root directory
            try:
                if (
                    _verifyFolderExists(target, destination)
                    and sanitized_destination != filename
                ):
                    # destination is an existing folder and not ourselves (= display rename), we'll assume we are supposed
                    # to move filename to this folder under the same name
                    destination = fileManager.join_path(target, destination, name)

                if _verifyFileExists(target, destination) or _verifyFolderExists(
                    target, destination
                ):
                    abort(409, description="File or folder does already exist")

            except Exception:
                abort(
                    409, description="Exception thrown by storage, bad folder/file name?"
                )

            is_file = fileManager.file_exists(target, filename)
            is_folder = fileManager.folder_exists(target, filename)

            if not (is_file or is_folder):
                abort(400, description=f"Neither file nor folder, can't {command}")

            try:
                if command == "copy":
                    # destination already there? error...
                    if _verifyFileExists(target, destination) or _verifyFolderExists(
                        target, destination
                    ):
                        abort(409, description="File or folder does already exist")

                    if is_file:
                        fileManager.copy_file(target, filename, destination)
                    else:
                        fileManager.copy_folder(target, filename, destination)

                elif command == "move":
                    with Permissions.FILES_DELETE.require(403):
                        if _isBusy(target, filename):
                            abort(
                                409,
                                description="Trying to move a file or folder that is currently in use",
                            )

                        # destination already there AND not ourselves (= display rename)? error...
                        if (
                            _verifyFileExists(target, destination)
                            or _verifyFolderExists(target, destination)
                        ) and sanitized_destination != filename:
                            abort(409, description="File or folder does already exist")

                        # deselect the file if it's currently selected
                        currentOrigin, currentFilename = _getCurrentFile()
                        if currentFilename is not None and filename == currentFilename:
                            printer.unselect_file()

                        if is_file:
                            fileManager.move_file(target, filename, destination)
                        else:
                            fileManager.move_folder(target, filename, destination)

            except octoprint.filemanager.storage.StorageError as e:
                if e.code == octoprint.filemanager.storage.StorageError.INVALID_FILE:
                    abort(
                        415,
                        description=f"Could not {command} {filename} to {destination}, invalid type",
                    )
                else:
                    abort(
                        500,
                        description=f"Could not {command} {filename} to {destination}",
                    )

            location = url_for(
                ".readGcodeFile",
                target=target,
                filename=destination,
                _external=True,
            )
            result = {
                "name": name,
                "path": destination,
                "origin": FileDestinations.LOCAL,
                "refs": {"resource": location},
            }
            if is_file:
                result["refs"]["download"] = (
                    url_for("index", _external=True)
                    + "downloads/files/"
                    + target
                    + "/"
                    + urlquote(destination)
                )

            r = make_response(jsonify(result), 201)
            r.headers["Location"] = location
            return r

    return NO_CONTENT


@api.route("/files/<string:target>/<path:filename>", methods=["DELETE"])
@no_firstrun_access
@Permissions.FILES_DELETE.require(403)
def deleteGcodeFile(filename, target):
    if not _validate(target, filename):
        abort(404)

    if not _verifyFileExists(target, filename) and not _verifyFolderExists(
        target, filename
    ):
        abort(404)

    if target not in [FileDestinations.LOCAL, FileDestinations.SDCARD]:
        abort(404)

    if _verifyFileExists(target, filename):
        if _isBusy(target, filename):
            abort(409, description="Trying to delete a file that is currently in use")

        # deselect the file if it's currently selected
        currentOrigin, currentPath = _getCurrentFile()
        if (
            currentPath is not None
            and currentOrigin == target
            and filename == currentPath
        ):
            printer.unselect_file()

        # delete it
        if target == FileDestinations.SDCARD:
            printer.delete_sd_file(filename, tags={"source:api", "api:files.sd"})
        else:
            try:
                fileManager.remove_file(target, filename)
            except (OSError, StorageError) as e:
                _abortWithException(e)

    elif _verifyFolderExists(target, filename):
        if _isBusy(target, filename):
            abort(
                409,
                description="Trying to delete a folder that contains a file that is currently in use",
            )

        # deselect the file if it's currently selected
        currentOrigin, currentPath = _getCurrentFile()
        if (
            currentPath is not None
            and currentOrigin == target
            and fileManager.file_in_path(target, filename, currentPath)
        ):
            printer.unselect_file()

        # delete it
        try:
            fileManager.remove_folder(target, filename, recursive=True)
        except (OSError, StorageError) as e:
            _abortWithException(e)

    return NO_CONTENT


def _abortWithException(error):
    if type(error) is StorageError:
        logging.getLogger(__name__).error(f"{error}: {error.code}", exc_info=error.cause)
        if error.code == StorageError.INVALID_DIRECTORY:
            abort(400, description="Could not create folder, invalid directory")
        elif error.code == StorageError.INVALID_FILE:
            abort(415, description="Could not upload file, invalid type")
        elif error.code == StorageError.INVALID_SOURCE:
            abort(404, description="Source path does not exist, invalid source")
        elif error.code == StorageError.INVALID_DESTINATION:
            abort(400, description="Destination is invalid")
        elif error.code == StorageError.DOES_NOT_EXIST:
            abort(404, description="Does not exit")
        elif error.code == StorageError.ALREADY_EXISTS:
            abort(409, description="File or folder already exists")
        elif error.code == StorageError.SOURCE_EQUALS_DESTINATION:
            abort(400, description="Source and destination are the same folder")
        elif error.code == StorageError.NOT_EMPTY:
            abort(409, description="Folder is not empty")
        elif error.code == StorageError.UNKNOWN:
            abort(500, description=str(error.cause).split(":")[0])
        else:
            abort(500, description=error)
    else:
        logging.getLogger(__name__).exception(error)
        abort(500, description=str(error).split(":")[0])


def _getCurrentFile():
    currentJob = printer.get_current_job()
    if (
        currentJob is not None
        and "file" in currentJob
        and "path" in currentJob["file"]
        and "origin" in currentJob["file"]
    ):
        return currentJob["file"]["origin"], currentJob["file"]["path"]
    else:
        return None, None


def _validate(target, filename):
    if target == FileDestinations.SDCARD:
        # we make no assumptions about the shape of valid SDCard file names
        return True
    else:
        return filename == "/".join(
            map(lambda x: fileManager.sanitize_name(target, x), filename.split("/"))
        )


class WerkzeugFileWrapper(octoprint.filemanager.util.AbstractFileWrapper):
    """
    A wrapper around a Werkzeug ``FileStorage`` object.

    Arguments:
        file_obj (werkzeug.datastructures.FileStorage): The Werkzeug ``FileStorage`` instance to wrap.

    .. seealso::

       `werkzeug.datastructures.FileStorage <http://werkzeug.pocoo.org/docs/0.10/datastructures/#werkzeug.datastructures.FileStorage>`_
            The documentation of Werkzeug's ``FileStorage`` class.
    """

    def __init__(self, file_obj):
        octoprint.filemanager.util.AbstractFileWrapper.__init__(self, file_obj.filename)
        self.file_obj = file_obj

    def save(self, path):
        """
        Delegates to ``werkzeug.datastructures.FileStorage.save``
        """
        self.file_obj.save(path)

    def stream(self):
        """
        Returns ``werkzeug.datastructures.FileStorage.stream``
        """
        return self.file_obj.stream



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

from enum import Enum
from typing import List, Optional

from octoprint.schema import BaseModel
from octoprint.vendor.with_attrs_docs import with_attrs_docs


class RatioEnum(str, Enum):
    sixteen_nine = "16:9"
    four_three = "4:3"


@with_attrs_docs
class WebcamCompatibility(BaseModel):
    streamTimeout: int = 5
    """The timeout of the stream in seconds"""

    streamRatio: RatioEnum = RatioEnum.sixteen_nine
    """The stream's native aspect ratio"""

    streamWebrtcIceServers: List[str] = ["stun:stun.l.google.com:19302"]
    """The WebRTC STUN and TURN servers"""

    cacheBuster: bool = False
    """Whether the the URL should be randomized to bust caches"""

    stream: str
    """The URL to get an MJPEG stream from"""

    snapshot: str = None
    """The URL to get the snapshot from"""

    snapshotTimeout: int = 5
    """The timeout when retrieving snapshots"""

    snapshotSslValidation: bool = True
    """Whether to validate SSL certificates when retrieving a snapshot"""


@with_attrs_docs
class Webcam(BaseModel):
    name: str
    """Identifier of this webcam"""

    displayName: str
    """Displayable name for this webcam"""

    canSnapshot: bool = False
    """Whether this webcam can take a snapshot."""

    snapshotDisplay: str = None
    """Human readable information about how a snapshot is captured or a HTTP URL from which the snapshot is loaded (optional, only for user reference)"""

    flipH: bool = False
    """Whether to flip the webcam horizontally."""

    flipV: bool = False
    """Whether to flip the webcam vertically."""

    rotate90: bool = False
    """Whether to rotate the webcam 90° counter clockwise."""

    extras: dict = {}
    """Unstructured data describing this webcam"""

    compat: Optional[WebcamCompatibility] = None
    """A compatibility configuration to allow older clients to make use of this webcam"""



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

import base64
import datetime
import json
import time
from collections import OrderedDict
from typing import Any, Callable, Dict, List

try:
    from typing import OrderedDict as OrderedDictType
except ImportError:
    OrderedDictType = Dict  # py3.7.{0,1}

from frozendict import frozendict

from .encoding import JsonEncoding


class SerializableJsonEncoding(JsonEncoding):
    """
    A JSON encoding that can serialize and deserialize objects, including additional
    objects otherwise not serializable by the standard JSON encoder:

      * ``bytes``
      * ``frozendict.frozendict``
      * ``datetime.datetime``
      * ``time.struct_time``
    """

    encoders: OrderedDictType[type, Callable[[Any], Any]] = OrderedDict()
    decoders: OrderedDictType[str, Callable[[dict], object]] = OrderedDict()

    @classmethod
    def add_decoder(cls, classname, decoder):
        cls.decoders[classname] = decoder

    @classmethod
    def remove_decoder(cls, classname):
        try:
            del cls.decoders[classname]
        except KeyError:
            pass

    @classmethod
    def dumps(cls, obj: Any) -> str:
        return json.dumps(
            cls.encode(obj), separators=(",", ":"), indent=None, allow_nan=False
        )

    @classmethod
    def loads(cls, s: str) -> Any:
        return json.loads(s, object_hook=cls.decode)

    @classmethod
    def encode(cls, val):
        """
        Recursively replace all instances of encodable types with their encoded
        value. This is useful over the ``default=`` functionality of the JSON
        encoder because JSON will not call default for tuples, lists, ints, etc:
        https://docs.python.org/3/library/json.html#json.JSONEncoder

        Cannot handle circular references.
        """

        if isinstance(val, tuple(cls.encoders.keys())):
            # we can't directly index into the encoders dict because
            # we need to be able to handle subclasses
            encoder = next(
                encoder for typ, encoder in cls.encoders.items() if isinstance(val, typ)
            )
            return cls.encode(encoder(val))
        elif isinstance(val, dict):
            return {k: cls.encode(v) for k, v in val.items()}
        elif isinstance(val, (tuple, list)):
            return [cls.encode(v) for v in val]
        elif isinstance(val, (bool, int, float, str)) or val is None:
            return val
        else:
            raise TypeError(f"Unserializable type {type(val)}")

    @classmethod
    def decode(cls, dct):
        """
        Recursively replace all instances of decodable types with their decoded
        values.

        You'll want to have used ``class_encode()`` in your encoder to get this
        to work properly.
        """
        if "__jsonclass__" not in dct:
            return dct

        if len(dct["__jsonclass__"]) == 0:
            raise ValueError("__jsonclass__ must not be empty")

        decoded_name = dct["__jsonclass__"][0]
        params = dct["__jsonclass__"][1:]

        for classname, decoder in cls.decoders.items():
            if decoded_name == classname:
                return decoder(*params)
        return dct


def class_encode(name: str, *params: Any) -> Dict[str, List]:
    """
    Encode a class name and parameters into a serializable dict. You'll
    probably want to use this if you're going to set a custom decoder.

    This stores the class names in a format inspired by the JSON-RPC spec at
    https://www.jsonrpc.org/specification_v1#a3.JSONClasshinting
    """
    return {"__jsonclass__": [name] + list(params)}


# frozendict

SerializableJsonEncoding.add_encoder(
    frozendict, lambda obj: class_encode("frozendict.frozendict", dict(obj))
)
SerializableJsonEncoding.add_decoder("frozendict.frozendict", lambda obj: frozendict(obj))

# bytes

SerializableJsonEncoding.add_encoder(
    bytes, lambda obj: class_encode("bytes", base64.b85encode(obj).decode("ascii"))
)
SerializableJsonEncoding.add_decoder("bytes", lambda obj: base64.b85decode(obj))

# time.struct_time


def _struct_time_decoder(params):
    if len(params) == 9 and all(isinstance(p, int) and p >= 0 for p in params):
        return time.struct_time(params)
    raise ValueError(f"Invalid time.struct_time params `{params}`")


SerializableJsonEncoding.add_encoder(
    time.struct_time, lambda obj: class_encode("time.struct_time", list(obj))
)
SerializableJsonEncoding.add_decoder("time.struct_time", _struct_time_decoder)

# datetime.datetime

SerializableJsonEncoding.add_encoder(
    datetime.datetime, lambda obj: class_encode("datetime.datetime", obj.isoformat())
)
SerializableJsonEncoding.add_decoder(
    "datetime.datetime", lambda params: datetime.datetime.fromisoformat(params)
)

# shortcut for dumps and loads

dumps = SerializableJsonEncoding.dumps
loads = SerializableJsonEncoding.loads



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

from typing import Optional

from octoprint.schema import BaseModel
from octoprint.vendor.with_attrs_docs import with_attrs_docs


@with_attrs_docs
class PrinterProfilesConfig(BaseModel):
    default: Optional[str] = None
    """Name of the printer profile to default to."""



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"


import json
import logging
import pprint

import click

from octoprint import FatalStartupError, init_settings
from octoprint.cli import get_ctx_obj_option, standard_options
from octoprint.util import yaml

click.disable_unicode_literals_warning = True


def _to_settings_path(path):
    if not isinstance(path, (list, tuple)):
        path = list(filter(lambda x: x, map(lambda x: x.strip(), path.split("."))))
    return path


def _set_helper(settings, path, value, data_type=None):
    path = _to_settings_path(path)

    method = settings.set
    if data_type is not None:
        name = None
        if data_type == bool:
            name = "setBoolean"
        elif data_type == float:
            name = "setFloat"
        elif data_type == int:
            name = "setInt"

        if name is not None:
            method = getattr(settings, name)

    method(path, value, force=True)
    settings.save()


# ~~ "octoprint config" commands


@click.group()
@click.pass_context
def cli(ctx):
    """Basic config manipulation."""
    logging.basicConfig(
        level=logging.DEBUG
        if get_ctx_obj_option(ctx, "verbosity", 0) > 0
        else logging.WARN
    )
    try:
        ctx.obj.settings = init_settings(
            get_ctx_obj_option(ctx, "basedir", None),
            get_ctx_obj_option(ctx, "configfile", None),
            overlays=get_ctx_obj_option(ctx, "overlays", None),
        )
    except FatalStartupError as e:
        click.echo(str(e), err=True)
        click.echo("There was a fatal error initializing the settings manager.", err=True)
        ctx.exit(-1)


@cli.command(name="set")
@standard_options(hidden=True)
@click.argument("path", type=click.STRING)
@click.argument("value", type=click.STRING)
@click.option("--bool", "as_bool", is_flag=True, help="Interpret value as bool")
@click.option("--float", "as_float", is_flag=True, help="Interpret value as float")
@click.option("--int", "as_int", is_flag=True, help="Interpret value as int")
@click.option("--json", "as_json", is_flag=True, help="Parse value from json")
@click.pass_context
def set_command(ctx, path, value, as_bool, as_float, as_int, as_json):
    """Sets a config path to the provided value."""
    if as_json:
        try:
            value = json.loads(value)
        except Exception as e:
            click.echo(str(e), err=True)
            ctx.exit(-1)

    data_type = None
    if as_bool:
        data_type = bool
    elif as_float:
        data_type = float
    elif as_int:
        data_type = int

    _set_helper(ctx.obj.settings, path, value, data_type=data_type)


@cli.command(name="remove")
@standard_options(hidden=True)
@click.argument("path", type=click.STRING)
@click.pass_context
def remove_command(ctx, path):
    """Removes a config path."""
    _set_helper(ctx.obj.settings, path, None)


@cli.command(name="append_value")
@standard_options(hidden=True)
@click.argument("path", type=click.STRING)
@click.argument("value", type=click.STRING)
@click.option("--json", "as_json", is_flag=True)
@click.pass_context
def append_value_command(ctx, path, value, as_json=False):
    """Appends value to list behind config path."""
    path = _to_settings_path(path)

    if as_json:
        try:
            value = json.loads(value)
        except Exception as e:
            click.echo(str(e), err=True)
            ctx.exit(-1)

    current = ctx.obj.settings.get(path)
    if current is None:
        current = []
    if not isinstance(current, list):
        click.echo("Cannot append to non-list value at given path", err=True)
        ctx.exit(-1)

    current.append(value)
    _set_helper(ctx.obj.settings, path, current)


@cli.command(name="insert_value")
@standard_options(hidden=True)
@click.argument("path", type=click.STRING)
@click.argument("index", type=click.INT)
@click.argument("value", type=click.STRING)
@click.option("--json", "as_json", is_flag=True)
@click.pass_context
def insert_value_command(ctx, path, index, value, as_json=False):
    """Inserts value at index of list behind config key."""
    path = _to_settings_path(path)

    if as_json:
        try:
            value = json.loads(value)
        except Exception as e:
            click.echo(str(e), err=True)
            ctx.exit(-1)

    current = ctx.obj.settings.get(path)
    if current is None:
        current = []
    if not isinstance(current, list):
        click.echo("Cannot insert into non-list value at given path", err=True)
        ctx.exit(-1)

    current.insert(index, value)
    _set_helper(ctx.obj.settings, path, current)


@cli.command(name="remove_value")
@standard_options(hidden=True)
@click.argument("path", type=click.STRING)
@click.argument("value", type=click.STRING)
@click.option("--json", "as_json", is_flag=True)
@click.pass_context
def remove_value_command(ctx, path, value, as_json=False):
    """Removes value from list at config path."""
    path = _to_settings_path(path)

    if as_json:
        try:
            value = json.loads(value)
        except Exception as e:
            click.echo(str(e), err=True)
            ctx.exit(-1)

    current = ctx.obj.settings.get(path)
    if current is None:
        current = []
    if not isinstance(current, list):
        click.echo("Cannot remove value from non-list value at given path", err=True)
        ctx.exit(-1)

    if value not in current:
        click.echo("Value is not contained in list at given path")
        ctx.exit()

    current.remove(value)
    _set_helper(ctx.obj.settings, path, current)


@cli.command(name="get")
@click.argument("path", type=click.STRING)
@click.option("--json", "as_json", is_flag=True, help="Output value formatted as JSON")
@click.option("--yaml", "as_yaml", is_flag=True, help="Output value formatted as YAML")
@click.option(
    "--raw", "as_raw", is_flag=True, help="Output value as raw string representation"
)
@standard_options(hidden=True)
@click.pass_context
def get_command(ctx, path, as_json=False, as_yaml=False, as_raw=False):
    """Retrieves value from config path."""
    path = _to_settings_path(path)
    value = ctx.obj.settings.get(path, merged=True)

    if as_json:
        output = json.dumps(value)
    elif as_yaml:
        output = yaml.dump(value, pretty=True)
    elif as_raw:
        output = value
    else:
        output = pprint.pformat(value)

    click.echo(output)


@cli.command(name="effective")
@click.option("--json", "as_json", is_flag=True, help="Output value formatted as JSON")
@click.option("--yaml", "as_yaml", is_flag=True, help="Output value formatted as YAML")
@click.option(
    "--raw", "as_raw", is_flag=True, help="Output value as raw string representation"
)
@standard_options(hidden=True)
@click.pass_context
def effective_command(ctx, as_json=False, as_yaml=False, as_raw=False):
    """Retrieves the full effective config."""
    value = ctx.obj.settings.effective

    if as_json:
        output = json.dumps(value)
    elif as_yaml:
        output = yaml.dump(value, pretty=True)
    elif as_raw:
        output = value
    else:
        output = pprint.pformat(value)

    click.echo(output)



__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

from flask import abort, jsonify, make_response, request, url_for

from octoprint.access.permissions import Permissions
from octoprint.server import slicingManager
from octoprint.server.api import NO_CONTENT, api
from octoprint.server.util.flask import no_firstrun_access, with_revalidation_checking
from octoprint.settings import settings as s
from octoprint.settings import valid_boolean_trues
from octoprint.slicing import (
    CouldNotDeleteProfile,
    SlicerNotConfigured,
    UnknownProfile,
    UnknownSlicer,
)

_DATA_FORMAT_VERSION = "v2"


def _lastmodified(configured):
    if configured:
        slicers = slicingManager.configured_slicers
    else:
        slicers = slicingManager.registered_slicers

    lms = [0]
    for slicer in slicers:
        lms.append(slicingManager.profiles_last_modified(slicer))

    return max(lms)


def _etag(configured, lm=None):
    if lm is None:
        lm = _lastmodified(configured)

    import hashlib

    hash = hashlib.sha1()

    def hash_update(value):
        value = value.encode("utf-8")
        hash.update(value)

    hash_update(str(lm))

    if configured:
        slicers = slicingManager.configured_slicers
    else:
        slicers = slicingManager.registered_slicers

    default_slicer = s().get(["slicing", "defaultSlicer"])

    for slicer in sorted(slicers):
        slicer_impl = slicingManager.get_slicer(slicer, require_configured=False)
        hash_update(slicer)
        hash_update(str(slicer_impl.is_slicer_configured()))
        hash_update(str(slicer == default_slicer))

    hash_update(_DATA_FORMAT_VERSION)  # increment version if we change the API format

    return hash.hexdigest()


@api.route("/slicing", methods=["GET"])
@with_revalidation_checking(
    etag_factory=lambda lm=None: _etag(
        request.values.get("configured", "false") in valid_boolean_trues, lm=lm
    ),
    lastmodified_factory=lambda: _lastmodified(
        request.values.get("configured", "false") in valid_boolean_trues
    ),
    unless=lambda: request.values.get("force", "false") in valid_boolean_trues,
)
@Permissions.SLICE.require(403)
def slicingListAll():
    from octoprint.filemanager import get_extensions

    default_slicer = s().get(["slicing", "defaultSlicer"])

    if (
        "configured" in request.values
        and request.values["configured"] in valid_boolean_trues
    ):
        slicers = slicingManager.configured_slicers
    else:
        slicers = slicingManager.registered_slicers

    result = {}
    for slicer in slicers:
        try:
            slicer_impl = slicingManager.get_slicer(slicer, require_configured=False)

            extensions = set()
            for source_file_type in slicer_impl.get_slicer_properties().get(
                "source_file_types", ["model"]
            ):
                extensions = extensions.union(get_extensions(source_file_type))

            result[slicer] = {
                "key": slicer,
                "displayName": slicer_impl.get_slicer_properties().get("name", "n/a"),
                "sameDevice": slicer_impl.get_slicer_properties().get(
                    "same_device", True
                ),
                "default": default_slicer == slicer,
                "configured": slicer_impl.is_slicer_configured(),
                "profiles": _getSlicingProfilesData(slicer),
                "extensions": {
                    "source": list(extensions),
                    "destination": slicer_impl.get_slicer_properties().get(
                        "destination_extensions", ["gco", "gcode", "g"]
                    ),
                },
            }
        except (UnknownSlicer, SlicerNotConfigured):
            # this should never happen
            pass

    return jsonify(result)


@api.route("/slicing/<string:slicer>/profiles", methods=["GET"])
@no_firstrun_access
@Permissions.SLICE.require(403)
def slicingListSlicerProfiles(slicer):
    configured = False
    if (
        "configured" in request.values
        and request.values["configured"] in valid_boolean_trues
    ):
        configured = True

    try:
        return jsonify(_getSlicingProfilesData(slicer, require_configured=configured))
    except (UnknownSlicer, SlicerNotConfigured):
        abort(404)


@api.route("/slicing/<string:slicer>/profiles/<string:name>", methods=["GET"])
@no_firstrun_access
@Permissions.SLICE.require(403)
def slicingGetSlicerProfile(slicer, name):
    try:
        profile = slicingManager.load_profile(slicer, name, require_configured=False)
    except UnknownSlicer:
        abort(404)
    except UnknownProfile:
        abort(404)

    result = _getSlicingProfileData(slicer, name, profile)
    result["data"] = profile.data
    return jsonify(result)


@api.route("/slicing/<string:slicer>/profiles/<string:name>", methods=["PUT"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def slicingAddSlicerProfile(slicer, name):
    json_data = request.get_json()

    data = {}
    display_name = None
    description = None
    if "data" in json_data:
        data = json_data["data"]
    if "displayName" in json_data:
        display_name = json_data["displayName"]
    if "description" in json_data:
        description = json_data["description"]

    try:
        profile = slicingManager.save_profile(
            slicer,
            name,
            data,
            allow_overwrite=True,
            display_name=display_name,
            description=description,
        )
    except UnknownSlicer:
        abort(404, description="Unknown slicer")

    result = _getSlicingProfileData(slicer, name, profile)
    r = make_response(jsonify(result), 201)
    r.headers["Location"] = result["resource"]
    return r


@api.route("/slicing/<string:slicer>/profiles/<string:name>", methods=["PATCH"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def slicingPatchSlicerProfile(slicer, name):
    json_data = request.get_json()

    try:
        profile = slicingManager.load_profile(slicer, name, require_configured=False)
    except UnknownSlicer:
        return abort(404)
    except UnknownProfile:
        return abort(404)

    data = {}
    display_name = None
    description = None
    if "data" in json_data:
        data = json_data["data"]
    if "displayName" in json_data:
        display_name = json_data["displayName"]
    if "description" in json_data:
        description = json_data["description"]

    saved_profile = slicingManager.save_profile(
        slicer,
        name,
        profile,
        allow_overwrite=True,
        overrides=data,
        display_name=display_name,
        description=description,
    )

    from octoprint.server.api import valid_boolean_trues

    if "default" in json_data and json_data["default"] in valid_boolean_trues:
        slicingManager.set_default_profile(slicer, name, require_exists=False)

    return jsonify(_getSlicingProfileData(slicer, name, saved_profile))


@api.route("/slicing/<string:slicer>/profiles/<string:name>", methods=["DELETE"])
@no_firstrun_access
@Permissions.SETTINGS.require(403)
def slicingDelSlicerProfile(slicer, name):
    try:
        slicingManager.delete_profile(slicer, name)
    except UnknownSlicer:
        abort(404)
    except CouldNotDeleteProfile as e:
        abort(
            500,
            description="Could not delete profile for slicer: {cause}".format(
                cause=str(e.cause)
            ),
        )

    return NO_CONTENT


def _getSlicingProfilesData(slicer, require_configured=False):
    profiles = slicingManager.all_profiles(slicer, require_configured=require_configured)

    result = {}
    for name, profile in profiles.items():
        result[name] = _getSlicingProfileData(slicer, name, profile)
    return result


def _getSlicingProfileData(slicer, name, profile):
    defaultProfiles = s().get(["slicing", "defaultProfiles"])
    result = {
        "key": name,
        "default": defaultProfiles
        and slicer in defaultProfiles
        and defaultProfiles[slicer] == name,
        "resource": url_for(
            ".slicingGetSlicerProfile", slicer=slicer, name=name, _external=True
        ),
    }
    if profile.display_name is not None:
        result["displayName"] = profile.display_name
    if profile.description is not None:
        result["description"] = profile.description
    return result



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

from typing import List

from octoprint.schema import BaseModel
from octoprint.vendor.with_attrs_docs import with_attrs_docs


@with_attrs_docs
class PrinterParametersConfig(BaseModel):
    pauseTriggers: List[str] = []



__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import logging
import os
import time
from collections import namedtuple

import octoprint.plugin
import octoprint.util
from octoprint.events import Events, eventManager
from octoprint.util import get_fully_qualified_classname as fqcn
from octoprint.util import yaml

from .analysis import AnalysisQueue, QueueEntry  # noqa: F401
from .destinations import FileDestinations  # noqa: F401
from .storage import LocalFileStorage  # noqa: F401
from .util import AbstractFileWrapper, DiskFileWrapper, StreamWrapper  # noqa: F401

ContentTypeMapping = namedtuple("ContentTypeMapping", "extensions, content_type")
ContentTypeDetector = namedtuple("ContentTypeDetector", "extensions, detector")

extensions = {}


def full_extension_tree():
    result = {
        "machinecode": {"gcode": ContentTypeMapping(["gcode", "gco", "g"], "text/plain")}
    }

    def leaf_merger(a, b):
        supported_leaf_types = (ContentTypeMapping, ContentTypeDetector, list)
        if not isinstance(a, supported_leaf_types) or not isinstance(
            b, supported_leaf_types
        ):
            raise ValueError()

        if isinstance(a, ContentTypeDetector) and isinstance(b, ContentTypeMapping):
            raise ValueError()

        if isinstance(a, ContentTypeMapping) and isinstance(b, ContentTypeDetector):
            raise ValueError()

        a_list = a if isinstance(a, list) else a.extensions
        b_list = b if isinstance(b, list) else b.extensions
        merged = a_list + b_list

        content_type = None
        if isinstance(b, ContentTypeMapping):
            content_type = b.content_type
        elif isinstance(a, ContentTypeMapping):
            content_type = a.content_type

        detector = None
        if isinstance(b, ContentTypeDetector):
            detector = b.detector
        elif isinstance(a, ContentTypeDetector):
            detector = a.detector

        if content_type is not None:
            return ContentTypeMapping(merged, content_type)
        elif detector is not None:
            return ContentTypeDetector(merged, detector)
        else:
            return merged

    slicer_plugins = octoprint.plugin.plugin_manager().get_implementations(
        octoprint.plugin.SlicerPlugin
    )
    for plugin in slicer_plugins:
        try:
            plugin_result = plugin.get_slicer_extension_tree()
            if plugin_result is None or not isinstance(plugin_result, dict):
                continue
            octoprint.util.dict_merge(
                result, plugin_result, leaf_merger=leaf_merger, in_place=True
            )
        except Exception:
            logging.getLogger(__name__).exception(
                "Exception while retrieving additional extension "
                "tree entries from SlicerPlugin {name}".format(name=plugin._identifier),
                extra={"plugin": plugin._identifier},
            )

    extension_tree_hooks = octoprint.plugin.plugin_manager().get_hooks(
        "octoprint.filemanager.extension_tree"
    )
    for name, hook in extension_tree_hooks.items():
        try:
            hook_result = hook()
            if hook_result is None or not isinstance(hook_result, dict):
                continue
            result = octoprint.util.dict_merge(
                result, hook_result, leaf_merger=leaf_merger, in_place=True
            )
        except Exception:
            logging.getLogger(__name__).exception(
                "Exception while retrieving additional extension "
                "tree entries from hook {name}".format(name=name),
                extra={"plugin": name},
            )

    return result


def get_extensions(type, subtree=None):
    if subtree is None:
        subtree = full_extension_tree()

    for key, value in subtree.items():
        if key == type:
            return get_all_extensions(subtree=value)
        elif isinstance(value, dict):
            sub_extensions = get_extensions(type, subtree=value)
            if sub_extensions:
                return sub_extensions

    return None


def get_all_extensions(subtree=None):
    if subtree is None:
        subtree = full_extension_tree()

    result = []
    if isinstance(subtree, dict):
        for value in subtree.values():
            if isinstance(value, dict):
                result += get_all_extensions(value)
            elif isinstance(value, (ContentTypeMapping, ContentTypeDetector)):
                result += value.extensions
            elif isinstance(value, (list, tuple)):
                result += value
    elif isinstance(subtree, (ContentTypeMapping, ContentTypeDetector)):
        result = subtree.extensions
    elif isinstance(subtree, (list, tuple)):
        result = subtree
    return result


def get_path_for_extension(extension, subtree=None):
    if subtree is None:
        subtree = full_extension_tree()

    for key, value in subtree.items():
        if (
            isinstance(value, (ContentTypeMapping, ContentTypeDetector))
            and extension in value.extensions
        ):
            return [key]
        elif isinstance(value, (list, tuple)) and extension in value:
            return [key]
        elif isinstance(value, dict):
            path = get_path_for_extension(extension, subtree=value)
            if path:
                return [key] + path

    return None


def get_content_type_mapping_for_extension(extension, subtree=None):
    if subtree is None:
        subtree = full_extension_tree()

    for value in subtree.values():
        content_extension_matches = (
            isinstance(value, (ContentTypeMapping, ContentTypeDetector))
            and extension in value.extensions
        )
        list_extension_matches = isinstance(value, (list, tuple)) and extension in value

        if content_extension_matches or list_extension_matches:
            return value
        elif isinstance(value, dict):
            result = get_content_type_mapping_for_extension(extension, subtree=value)
            if result is not None:
                return result

    return None


def valid_extension(extension, type=None, tree=None):
    if not type:
        return extension in get_all_extensions(subtree=tree)
    else:
        extensions = get_extensions(type, subtree=tree)
        if extensions:
            return extension in extensions


def valid_file_type(filename, type=None, tree=None):
    _, extension = os.path.splitext(filename)
    extension = extension[1:].lower()
    return valid_extension(extension, type=type, tree=tree)


def get_file_type(filename):
    _, extension = os.path.splitext(filename)
    extension = extension[1:].lower()
    return get_path_for_extension(extension)


def get_mime_type(filename):
    _, extension = os.path.splitext(filename)
    extension = extension[1:].lower()
    mapping = get_content_type_mapping_for_extension(extension)
    if mapping:
        if isinstance(mapping, ContentTypeMapping) and mapping.content_type is not None:
            return mapping.content_type
        elif isinstance(mapping, ContentTypeDetector) and callable(mapping.detector):
            result = mapping.detector(filename)
            if result is not None:
                return result
    return "application/octet-stream"


class NoSuchStorage(Exception):
    pass


class FileManager:
    def __init__(
        self,
        analysis_queue,
        slicing_manager,
        printer_profile_manager,
        initial_storage_managers=None,
    ):
        self._logger = logging.getLogger(__name__)
        self._analysis_queue = analysis_queue
        self._analysis_queue.register_finish_callback(self._on_analysis_finished)

        self._storage_managers = {}
        if initial_storage_managers:
            self._storage_managers.update(initial_storage_managers)

        self._slicing_manager = slicing_manager
        self._printer_profile_manager = printer_profile_manager

        import threading

        self._slicing_jobs = {}
        self._slicing_jobs_mutex = threading.Lock()

        self._slicing_progress_callbacks = []
        self._last_slicing_progress = None

        self._progress_plugins = []
        self._preprocessor_hooks = {}

        import octoprint.settings

        self._recovery_file = os.path.join(
            octoprint.settings.settings().getBaseFolder("data"),
            "print_recovery_data.yaml",
        )
        self._analyzeGcode = octoprint.settings.settings().get(["gcodeAnalysis", "runAt"])

    def initialize(self, process_backlog=False):
        self.reload_plugins()
        if process_backlog:
            self.process_backlog()

    def process_backlog(self):
        # only check for a backlog if gcodeAnalysis is 'idle' or 'always'
        if self._analyzeGcode == "never":
            return

        def worker():
            self._logger.info(
                "Adding backlog items from all storage types to analysis queue..."
            )
            for storage_type, storage_manager in self._storage_managers.items():
                self._determine_analysis_backlog(storage_type, storage_manager)

        import threading

        thread = threading.Thread(target=worker)
        thread.daemon = True
        thread.start()

    def reload_plugins(self):
        self._progress_plugins = octoprint.plugin.plugin_manager().get_implementations(
            octoprint.plugin.ProgressPlugin
        )
        self._preprocessor_hooks = octoprint.plugin.plugin_manager().get_hooks(
            "octoprint.filemanager.preprocessor"
        )

    def register_slicingprogress_callback(self, callback):
        self._slicing_progress_callbacks.append(callback)

    def unregister_slicingprogress_callback(self, callback):
        try:
            self._slicing_progress_callbacks.remove(callback)
        except ValueError:
            # callback was not registered
            pass

    def _determine_analysis_backlog(
        self, storage_type, storage_manager, root=None, high_priority=False
    ):
        counter = 0

        backlog_generator = storage_manager.analysis_backlog
        if root is not None:
            backlog_generator = storage_manager.analysis_backlog_for_path(path=root)

        for entry, path, _ in backlog_generator:
            file_type = get_file_type(path)[-1]
            file_name = storage_manager.split_path(path)

            # we'll use the default printer profile for the backlog since we don't know better
            queue_entry = QueueEntry(
                file_name,
                entry,
                file_type,
                storage_type,
                path,
                self._printer_profile_manager.get_default(),
                None,
            )
            if self._analysis_queue.enqueue(queue_entry, high_priority=high_priority):
                counter += 1

        if root:
            self._logger.info(
                f'Added {counter} items from storage type "{storage_type}" and root "{root}" to analysis queue'
            )
        else:
            self._logger.info(
                f'Added {counter} items from storage type "{storage_type}" to analysis queue'
            )

    def add_storage(self, storage_type, storage_manager):
        self._storage_managers[storage_type] = storage_manager
        self._determine_analysis_backlog(storage_type, storage_manager)

    def remove_storage(self, type):
        if type not in self._storage_managers:
            return
        del self._storage_managers[type]

    @property
    def registered_storages(self):
        return list(self._storage_managers.keys())

    @property
    def slicing_enabled(self):
        return self._slicing_manager.slicing_enabled

    @property
    def registered_slicers(self):
        return self._slicing_manager.registered_slicers

    @property
    def default_slicer(self):
        return self._slicing_manager.default_slicer

    def analyse(self, destination, path, printer_profile_id=None):
        if not self.file_exists(destination, path):
            return

        if printer_profile_id is None:
            printer_profile = self._printer_profile_manager.get_current_or_default()
        else:
            printer_profile = self._printer_profile_manager.get(printer_profile_id)
            if printer_profile is None:
                printer_profile = self._printer_profile_manager.get_current_or_default()

        queue_entry = self._analysis_queue_entry(destination, path)
        self._analysis_queue.dequeue(queue_entry)

        queue_entry = self._analysis_queue_entry(
            destination, path, printer_profile=printer_profile
        )
        if queue_entry:
            return self._analysis_queue.enqueue(queue_entry, high_priority=True)

        return False

    def slice(
        self,
        slicer_name,
        source_location,
        source_path,
        dest_location,
        dest_path,
        position=None,
        profile=None,
        printer_profile_id=None,
        overrides=None,
        display=None,
        callback=None,
        callback_args=None,
    ):
        absolute_source_path = self.path_on_disk(source_location, source_path)

        def stlProcessed(
            source_location,
            source_path,
            tmp_path,
            dest_location,
            dest_path,
            start_time,
            printer_profile_id,
            callback,
            callback_args,
            _error=None,
            _cancelled=False,
            _analysis=None,
        ):
            try:
                if _error:
                    eventManager().fire(
                        Events.SLICING_FAILED,
                        {
                            "slicer": slicer_name,
                            "stl": source_path,
                            "stl_location": source_location,
                            "gcode": dest_path,
                            "gcode_location": dest_location,
                            "reason": _error,
                        },
                    )
                elif _cancelled:
                    eventManager().fire(
                        Events.SLICING_CANCELLED,
                        {
                            "slicer": slicer_name,
                            "stl": source_path,
                            "stl_location": source_location,
                            "gcode": dest_path,
                            "gcode_location": dest_location,
                        },
                    )
                else:
                    source_meta = self.get_metadata(source_location, source_path)
                    hash = source_meta.get("hash", "n/a")

                    import io

                    links = [("model", {"name": source_path})]
                    _, stl_name = self.split_path(source_location, source_path)
                    file_obj = StreamWrapper(
                        os.path.basename(dest_path),
                        io.BytesIO(
                            f";Generated from {stl_name} (hash: {hash})\n".encode(
                                "ascii", "replace"
                            )
                        ),
                        io.FileIO(tmp_path, "rb"),
                    )

                    printer_profile = self._printer_profile_manager.get(
                        printer_profile_id
                    )
                    self.add_file(
                        dest_location,
                        dest_path,
                        file_obj,
                        display=display,
                        links=links,
                        allow_overwrite=True,
                        printer_profile=printer_profile,
                        analysis=_analysis,
                    )

                    end_time = time.monotonic()
                    eventManager().fire(
                        Events.SLICING_DONE,
                        {
                            "slicer": slicer_name,
                            "stl": source_path,
                            "stl_location": source_location,
                            "gcode": dest_path,
                            "gcode_location": dest_location,
                            "time": end_time - start_time,
                        },
                    )

                    if callback is not None:
                        if callback_args is None:
                            callback_args = ()
                        callback(*callback_args)
            finally:
                os.remove(tmp_path)

                source_job_key = (source_location, source_path)
                dest_job_key = (dest_location, dest_path)

                with self._slicing_jobs_mutex:
                    if source_job_key in self._slicing_jobs:
                        del self._slicing_jobs[source_job_key]
                    if dest_job_key in self._slicing_jobs:
                        del self._slicing_jobs[dest_job_key]

        slicer = self._slicing_manager.get_slicer(slicer_name)

        start_time = time.monotonic()
        eventManager().fire(
            Events.SLICING_STARTED,
            {
                "slicer": slicer_name,
                "stl": source_path,
                "stl_location": source_location,
                "gcode": dest_path,
                "gcode_location": dest_location,
                "progressAvailable": slicer.get_slicer_properties().get(
                    "progress_report", False
                )
                if slicer
                else False,
            },
        )

        import tempfile

        f = tempfile.NamedTemporaryFile(suffix=".gco", delete=False)
        temp_path = f.name
        f.close()

        with self._slicing_jobs_mutex:
            source_job_key = (source_location, source_path)
            dest_job_key = (dest_location, dest_path)
            if dest_job_key in self._slicing_jobs:
                (
                    job_slicer_name,
                    job_absolute_source_path,
                    job_temp_path,
                ) = self._slicing_jobs[dest_job_key]

                self._slicing_manager.cancel_slicing(
                    job_slicer_name, job_absolute_source_path, job_temp_path
                )
                del self._slicing_jobs[dest_job_key]

            self._slicing_jobs[dest_job_key] = self._slicing_jobs[source_job_key] = (
                slicer_name,
                absolute_source_path,
                temp_path,
            )

        args = (
            source_location,
            source_path,
            temp_path,
            dest_location,
            dest_path,
            start_time,
            printer_profile_id,
            callback,
            callback_args,
        )
        self._slicing_manager.slice(
            slicer_name,
            absolute_source_path,
            temp_path,
            profile,
            stlProcessed,
            position=position,
            callback_args=args,
            overrides=overrides,
            printer_profile_id=printer_profile_id,
            on_progress=self.on_slicing_progress,
            on_progress_args=(
                slicer_name,
                source_location,
                source_path,
                dest_location,
                dest_path,
            ),
        )

    def on_slicing_progress(
        self,
        slicer,
        source_location,
        source_path,
        dest_location,
        dest_path,
        _progress=None,
    ):
        if not _progress:
            return

        progress_int = int(_progress * 100)
        if self._last_slicing_progress != progress_int:
            self._last_slicing_progress = progress_int
            for callback in self._slicing_progress_callbacks:
                try:
                    callback.sendSlicingProgress(
                        slicer,
                        source_location,
                        source_path,
                        dest_location,
                        dest_path,
                        progress_int,
                    )
                except Exception:
                    self._logger.exception(
                        "Exception while pushing slicing progress",
                        extra={"callback": fqcn(callback)},
                    )

            if progress_int:

                def call_plugins(
                    slicer,
                    source_location,
                    source_path,
                    dest_location,
                    dest_path,
                    progress,
                ):
                    for plugin in self._progress_plugins:
                        try:
                            plugin.on_slicing_progress(
                                slicer,
                                source_location,
                                source_path,
                                dest_location,
                                dest_path,
                                progress,
                            )
                        except Exception:
                            self._logger.exception(
                                "Exception while sending slicing progress to plugin %s"
                                % plugin._identifier,
                                extra={"plugin": plugin._identifier},
                            )

                import threading

                thread = threading.Thread(
                    target=call_plugins,
                    args=(
                        slicer,
                        source_location,
                        source_path,
                        dest_location,
                        dest_path,
                        progress_int,
                    ),
                )
                thread.daemon = False
                thread.start()

    def get_busy_files(self):
        return self._slicing_jobs.keys()

    def file_in_path(self, destination, path, file):
        return self._storage(destination).file_in_path(path, file)

    def file_exists(self, destination, path):
        return self._storage(destination).file_exists(path)

    def folder_exists(self, destination, path):
        return self._storage(destination).folder_exists(path)

    def list_files(
        self,
        destinations=None,
        path=None,
        filter=None,
        recursive=None,
        level=0,
        force_refresh=False,
    ):
        if not destinations:
            destinations = list(self._storage_managers.keys())
        if isinstance(destinations, str):
            destinations = [destinations]

        result = {}
        for dst in destinations:
            result[dst] = self._storage_managers[dst].list_files(
                path=path,
                filter=filter,
                recursive=recursive,
                level=level,
                force_refresh=force_refresh,
            )
        return result

    def add_file(
        self,
        destination,
        path,
        file_object,
        links=None,
        allow_overwrite=False,
        printer_profile=None,
        analysis=None,
        display=None,
    ):
        if printer_profile is None:
            printer_profile = self._printer_profile_manager.get_current_or_default()

        for name, hook in self._preprocessor_hooks.items():
            try:
                hook_file_object = hook(
                    path,
                    file_object,
                    links=links,
                    printer_profile=printer_profile,
                    allow_overwrite=allow_overwrite,
                )
            except Exception:
                self._logger.exception(
                    "Error when calling preprocessor hook for plugin {}, ignoring".format(
                        name
                    ),
                    extra={"plugin": name},
                )
                continue

            if hook_file_object is not None:
                file_object = hook_file_object

        queue_entry = self._analysis_queue_entry(destination, path)

        if queue_entry is not None:
            self._analysis_queue.dequeue(queue_entry)

        path_in_storage = self._storage(destination).add_file(
            path,
            file_object,
            links=links,
            printer_profile=printer_profile,
            allow_overwrite=allow_overwrite,
            display=display,
        )

        queue_entry = self._analysis_queue_entry(
            destination,
            path_in_storage,
            printer_profile=printer_profile,
            analysis=analysis,
        )
        if queue_entry:
            self._analysis_queue.enqueue(queue_entry, high_priority=True)

        _, name = self._storage(destination).split_path(path_in_storage)
        eventManager().fire(
            Events.FILE_ADDED,
            {
                "storage": destination,
                "path": path_in_storage,
                "name": name,
                "type": get_file_type(name),
                "operation": "add",
            },
        )
        eventManager().fire(Events.UPDATED_FILES, {"type": "printables"})
        return path_in_storage

    def remove_file(self, destination, path):
        queue_entry = self._analysis_queue_entry(destination, path)
        self._analysis_queue.dequeue(queue_entry)
        self._storage(destination).remove_file(path)

        _, name = self._storage(destination).split_path(path)
        eventManager().fire(
            Events.FILE_REMOVED,
            {
                "storage": destination,
                "path": path,
                "name": name,
                "type": get_file_type(name),
            },
        )
        eventManager().fire(Events.UPDATED_FILES, {"type": "printables"})

    def copy_file(self, destination, source, dst):
        path_in_storage = self._storage(destination).copy_file(source, dst)
        if not self.has_analysis(destination, path_in_storage):
            queue_entry = self._analysis_queue_entry(destination, path_in_storage)
            if queue_entry:
                self._analysis_queue.enqueue(queue_entry)

        _, name = self._storage(destination).split_path(path_in_storage)
        eventManager().fire(
            Events.FILE_ADDED,
            {
                "storage": destination,
                "path": path_in_storage,
                "name": name,
                "type": get_file_type(name),
                "operation": "copy",
            },
        )
        eventManager().fire(Events.UPDATED_FILES, {"type": "printables"})

    def move_file(self, destination, source, dst):
        queue_entry = self._analysis_queue_entry(destination, source)
        self._analysis_queue.dequeue(queue_entry)
        path = self._storage(destination).move_file(source, dst)
        if not self.has_analysis(destination, path):
            queue_entry = self._analysis_queue_entry(destination, path)
            if queue_entry:
                self._analysis_queue.enqueue(queue_entry)

        source_path_in_storage = self._storage(destination).path_in_storage(source)
        _, source_name = self._storage(destination).split_path(source_path_in_storage)
        dst_path_in_storage = self._storage(destination).path_in_storage(dst)
        _, dst_name = self._storage(destination).split_path(dst_path_in_storage)

        eventManager().fire(
            Events.FILE_REMOVED,
            {
                "storage": destination,
                "path": source_path_in_storage,
                "name": source_name,
                "type": get_file_type(source_name),
            },
        )
        eventManager().fire(
            Events.FILE_ADDED,
            {
                "storage": destination,
                "path": dst_path_in_storage,
                "name": dst_name,
                "type": get_file_type(dst_name),
                "operation": "move",
            },
        )
        eventManager().fire(
            Events.FILE_MOVED,
            {
                "storage": destination,
                "source_path": source_path_in_storage,
                "source_name": source_name,
                "source_type": get_file_type(source_name),
                "destination_path": dst_path_in_storage,
                "destination_name": dst_name,
                "destination_type": get_file_type(dst_name),
            },
        )

        eventManager().fire(Events.UPDATED_FILES, {"type": "printables"})

    def add_folder(self, destination, path, ignore_existing=True, display=None):
        path_in_storage = self._storage(destination).add_folder(
            path, ignore_existing=ignore_existing, display=display
        )

        _, name = self._storage(destination).split_path(path_in_storage)
        eventManager().fire(
            Events.FOLDER_ADDED,
            {"storage": destination, "path": path_in_storage, "name": name},
        )
        eventManager().fire(Events.UPDATED_FILES, {"type": "printables"})
        return path_in_storage

    def remove_folder(self, destination, path, recursive=True):
        self._analysis_queue.dequeue_folder(destination, path)
        self._analysis_queue.pause()
        self._storage(destination).remove_folder(path, recursive=recursive)
        self._analysis_queue.resume()

        _, name = self._storage(destination).split_path(path)
        eventManager().fire(
            Events.FOLDER_REMOVED, {"storage": destination, "path": path, "name": name}
        )
        eventManager().fire(Events.UPDATED_FILES, {"type": "printables"})

    def copy_folder(self, destination, source, dst):
        path_in_storage = self._storage(destination).copy_folder(source, dst)
        self._determine_analysis_backlog(
            destination, self._storage(destination), root=path_in_storage
        )

        _, name = self._storage(destination).split_path(path_in_storage)
        eventManager().fire(
            Events.FOLDER_ADDED,
            {"storage": destination, "path": path_in_storage, "name": name},
        )
        eventManager().fire(Events.UPDATED_FILES, {"type": "printables"})

    def move_folder(self, destination, source, dst):
        self._analysis_queue.dequeue_folder(destination, source)
        self._analysis_queue.pause()
        dst_path_in_storage = self._storage(destination).move_folder(source, dst)
        self._determine_analysis_backlog(
            destination, self._storage(destination), root=dst_path_in_storage
        )
        self._analysis_queue.resume()

        source_path_in_storage = self._storage(destination).path_in_storage(source)
        _, source_name = self._storage(destination).split_path(source_path_in_storage)
        _, dst_name = self._storage(destination).split_path(dst_path_in_storage)

        eventManager().fire(
            Events.FOLDER_REMOVED,
            {"storage": destination, "path": source_path_in_storage, "name": source_name},
        )
        eventManager().fire(
            Events.FOLDER_ADDED,
            {"storage": destination, "path": dst_path_in_storage, "name": dst_name},
        )
        eventManager().fire(
            Events.FOLDER_MOVED,
            {
                "storage": destination,
                "source_path": source_path_in_storage,
                "source_name": source_name,
                "destination_path": dst_path_in_storage,
                "destination_name": dst_name,
            },
        )

        eventManager().fire(Events.UPDATED_FILES, {"type": "printables"})

    def has_analysis(self, destination, path):
        return self._storage(destination).has_analysis(path)

    def get_metadata(self, destination, path):
        return self._storage(destination).get_metadata(path)

    def add_link(self, destination, path, rel, data):
        self._storage(destination).add_link(path, rel, data)

    def remove_link(self, destination, path, rel, data):
        self._storage(destination).remove_link(path, rel, data)

    def log_print(
        self, destination, path, timestamp, print_time, success, printer_profile
    ):
        try:
            if success:
                self._storage(destination).add_history(
                    path,
                    {
                        "timestamp": timestamp,
                        "printTime": print_time,
                        "success": success,
                        "printerProfile": printer_profile,
                    },
                )
            else:
                self._storage(destination).add_history(
                    path,
                    {
                        "timestamp": timestamp,
                        "success": success,
                        "printerProfile": printer_profile,
                    },
                )
            eventManager().fire(
                Events.METADATA_STATISTICS_UPDATED, {"storage": destination, "path": path}
            )
        except NoSuchStorage:
            # if there's no storage configured where to log the print, we'll just not log it
            pass

    def save_recovery_data(self, origin, path, pos):
        import time

        from octoprint.util import atomic_write

        data = {
            "origin": origin,
            "path": self.path_in_storage(origin, path),
            "pos": pos,
            "date": time.time(),
        }
        try:
            with atomic_write(self._recovery_file, mode="wt", max_permissions=0o666) as f:
                yaml.save_to_file(data, file=f, pretty=True)
        except Exception:
            self._logger.exception(
                f"Could not write recovery data to file {self._recovery_file}"
            )

    def delete_recovery_data(self):
        if not os.path.isfile(self._recovery_file):
            return

        try:
            os.remove(self._recovery_file)
        except Exception:
            self._logger.exception(
                f"Error deleting recovery data file {self._recovery_file}"
            )

    def get_recovery_data(self):
        if not os.path.isfile(self._recovery_file):
            return None

        try:
            data = yaml.load_from_file(path=self._recovery_file)

            if not isinstance(data, dict) or not all(
                map(lambda x: x in data, ("origin", "path", "pos", "date"))
            ):
                raise ValueError("Invalid recovery data structure")
            return data
        except Exception:
            self._logger.exception(
                f"Could not read recovery data from file {self._recovery_file}"
            )
            self.delete_recovery_data()

    def get_additional_metadata(self, destination, path, key):
        self._storage(destination).get_additional_metadata(path, key)

    def set_additional_metadata(
        self, destination, path, key, data, overwrite=False, merge=False
    ):
        self._storage(destination).set_additional_metadata(
            path, key, data, overwrite=overwrite, merge=merge
        )

    def remove_additional_metadata(self, destination, path, key):
        self._storage(destination).remove_additional_metadata(path, key)

    def path_on_disk(self, destination, path):
        return self._storage(destination).path_on_disk(path)

    def canonicalize(self, destination, path):
        return self._storage(destination).canonicalize(path)

    def sanitize(self, destination, path):
        return self._storage(destination).sanitize(path)

    def sanitize_name(self, destination, name):
        return self._storage(destination).sanitize_name(name)

    def sanitize_path(self, destination, path):
        return self._storage(destination).sanitize_path(path)

    def split_path(self, destination, path):
        return self._storage(destination).split_path(path)

    def join_path(self, destination, *path):
        return self._storage(destination).join_path(*path)

    def path_in_storage(self, destination, path):
        return self._storage(destination).path_in_storage(path)

    def last_modified(self, destination, path=None, recursive=False):
        return self._storage(destination).last_modified(path=path, recursive=recursive)

    def _storage(self, destination):
        if destination not in self._storage_managers:
            raise NoSuchStorage(f"No storage configured for destination {destination}")
        return self._storage_managers[destination]

    def _add_analysis_result(self, destination, path, result):
        if destination not in self._storage_managers:
            return
        if not result:
            return

        storage_manager = self._storage_managers[destination]
        storage_manager.set_additional_metadata(path, "analysis", result, overwrite=True)

    def _on_analysis_finished(self, entry, result):
        self._add_analysis_result(entry.location, entry.path, result)

    def _analysis_queue_entry(
        self, destination, path, printer_profile=None, analysis=None
    ):
        if printer_profile is None:
            printer_profile = self._printer_profile_manager.get_current_or_default()

        absolute_path = self._storage(destination).path_on_disk(path)
        _, file_name = self._storage(destination).split_path(path)
        file_type = get_file_type(absolute_path)

        if file_type:
            return QueueEntry(
                file_name,
                path,
                file_type[-1],
                destination,
                absolute_path,
                printer_profile,
                analysis,
            )
        else:
            return None



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2020 The OctoPrint Project - Released under terms of the AGPLv3 License"


class InvalidPackageFormat(Exception):
    pass



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"

import json

import click

import octoprint_client
from octoprint import FatalStartupError, init_settings
from octoprint.cli import bulk_options, get_ctx_obj_option
from octoprint.util import yaml

click.disable_unicode_literals_warning = True


class JsonStringParamType(click.ParamType):
    name = "json"

    def convert(self, value, param, ctx):
        try:
            return json.loads(value)
        except Exception:
            self.fail("%s is not a valid json string" % value, param, ctx)


def create_client(
    settings=None,
    apikey=None,
    host=None,
    port=None,
    httpuser=None,
    httppass=None,
    https=False,
    prefix=None,
):
    assert host is not None or settings is not None
    assert port is not None or settings is not None
    assert apikey is not None or settings is not None

    if not host:
        host = settings.get(["server", "host"])
        host = host if host != "0.0.0.0" else "127.0.0.1"
    if not port:
        port = settings.getInt(["server", "port"])

    if not apikey:
        apikey = settings.get(["api", "key"])

    baseurl = octoprint_client.build_base_url(
        https=https,
        httpuser=httpuser,
        httppass=httppass,
        host=host,
        port=port,
        prefix=prefix,
    )

    return octoprint_client.Client(baseurl, apikey)


client_options = bulk_options(
    [
        click.option("--apikey", "-a", type=click.STRING),
        click.option("--host", "-h", type=click.STRING),
        click.option("--port", "-p", type=click.INT),
        click.option("--httpuser", type=click.STRING),
        click.option("--httppass", type=click.STRING),
        click.option("--https", is_flag=True),
        click.option("--prefix", type=click.STRING),
    ]
)
"""Common options to configure an API client."""


@click.group(context_settings={"ignore_unknown_options": True})
@client_options
@click.pass_context
def cli(ctx, apikey, host, port, httpuser, httppass, https, prefix):
    """Basic API client."""
    try:
        settings = None
        if not host or not port or not apikey:
            settings = init_settings(
                get_ctx_obj_option(ctx, "basedir", None),
                get_ctx_obj_option(ctx, "configfile", None),
            )

        ctx.obj.client = create_client(
            settings=settings,
            apikey=apikey,
            host=host,
            port=port,
            httpuser=httpuser,
            httppass=httppass,
            https=https,
            prefix=prefix,
        )

    except FatalStartupError as e:
        click.echo(str(e), err=True)
        click.echo("There was a fatal error initializing the client.", err=True)
        ctx.exit(-1)


def log_response(response, status_code=True, body=True, headers=False):
    if status_code:
        click.echo(f"Status Code: {response.status_code}")
    if headers:
        for header, value in response.headers.items():
            click.echo(f"{header}: {value}")
        click.echo()
    if body:
        click.echo(response.text)


@cli.command("get")
@click.argument("path")
@click.option("--timeout", type=float, default=None)
@click.pass_context
def get(ctx, path, timeout):
    """Performs a GET request against the specified server path."""
    r = ctx.obj.client.get(path, timeout=timeout)
    log_response(r)


@cli.command("post_json")
@click.argument("path")
@click.argument("data", type=JsonStringParamType())
@click.option("--timeout", type=float, default=None)
@click.pass_context
def post_json(ctx, path, data, timeout):
    """POSTs JSON data to the specified server path."""
    r = ctx.obj.client.post_json(path, data, timeout=timeout)
    log_response(r)


@cli.command("patch_json")
@click.argument("path")
@click.argument("data", type=JsonStringParamType())
@click.option("--timeout", type=float, default=None, help="Request timeout in seconds")
@click.pass_context
def patch_json(ctx, path, data, timeout):
    """PATCHes JSON data to the specified server path."""
    r = ctx.obj.client.patch(path, data, encoding="json", timeout=timeout)
    log_response(r)


@cli.command("post_from_file")
@click.argument("path")
@click.argument(
    "file_path", type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
@click.option("--json", is_flag=True)
@click.option("--yaml", is_flag=True)
@click.option("--timeout", type=float, default=None, help="Request timeout in seconds")
@click.pass_context
def post_from_file(ctx, path, file_path, json_flag, yaml_flag, timeout):
    """POSTs JSON data to the specified server path, taking the data from the specified file."""
    if json_flag or yaml_flag:
        if json_flag:
            with open(file_path) as fp:
                data = json.load(fp)
        else:
            data = yaml.load_from_file(path=file_path)

        r = ctx.obj.client.post_json(path, data, timeout=timeout)
    else:
        with open(file_path, "rb") as fp:
            data = fp.read()

        r = ctx.obj.client.post(path, data, timeout=timeout)

    log_response(r)


@cli.command("command")
@click.argument("path")
@click.argument("command")
@click.option(
    "--str",
    "-s",
    "str_params",
    multiple=True,
    nargs=2,
    type=click.Tuple([str, str]),
)
@click.option(
    "--int", "-i", "int_params", multiple=True, nargs=2, type=click.Tuple([str, int])
)
@click.option(
    "--float",
    "-f",
    "float_params",
    multiple=True,
    nargs=2,
    type=click.Tuple([str, float]),
)
@click.option(
    "--bool",
    "-b",
    "bool_params",
    multiple=True,
    nargs=2,
    type=click.Tuple([str, bool]),
)
@click.option("--timeout", type=float, default=None, help="Request timeout in seconds")
@click.pass_context
def command(
    ctx, path, command, str_params, int_params, float_params, bool_params, timeout
):
    """Sends a JSON command to the specified server path."""
    data = {}
    params = str_params + int_params + float_params + bool_params
    for param in params:
        data[param[0]] = param[1]
    r = ctx.obj.client.post_command(path, command, additional=data, timeout=timeout)
    log_response(r, body=False)


@cli.command("upload")
@click.argument("path")
@click.argument(
    "file_path", type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
@click.option(
    "--parameter",
    "-P",
    "params",
    multiple=True,
    nargs=2,
    type=click.Tuple([str, str]),
)
@click.option("--file-name", type=click.STRING)
@click.option("--content-type", type=click.STRING)
@click.option("--timeout", type=float, default=None, help="Request timeout in seconds")
@click.pass_context
def upload(ctx, path, file_path, params, file_name, content_type, timeout):
    """Uploads the specified file to the specified server path."""
    data = {}
    for param in params:
        data[param[0]] = param[1]

    r = ctx.obj.client.upload(
        path,
        file_path,
        additional=data,
        file_name=file_name,
        content_type=content_type,
        timeout=timeout,
    )
    log_response(r)


@cli.command("delete")
@click.argument("path")
@click.option("--timeout", type=float, default=None, help="Request timeout in seconds")
@click.pass_context
def delete(ctx, path, timeout):
    """Sends a DELETE request to the specified server path."""
    r = ctx.obj.client.delete(path, timeout=timeout)
    log_response(r)


@cli.command("listen")
@click.pass_context
def listen(ctx):
    def on_connect(ws):
        click.echo("--- Connected!")

    def on_close(ws):
        click.echo("--- Connection closed!")

    def on_error(ws, error):
        click.echo(f"!!! Error: {error}")

    def on_sent(ws, data):
        click.echo(f">>> {json.dumps(data)}")

    def on_heartbeat(ws):
        click.echo("<3")

    def on_message(ws, message_type, message_payload):
        click.echo(f"<<< {message_type}, Payload: {json.dumps(message_payload)}")

    socket = ctx.obj.client.create_socket(
        on_connect=on_connect,
        on_close=on_close,
        on_error=on_error,
        on_sent=on_sent,
        on_heartbeat=on_heartbeat,
        on_message=on_message,
    )

    click.echo("--- Waiting for client to exit")
    try:
        socket.wait()
    finally:
        click.echo("--- Goodbye...")



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"


import sys

import click

import octoprint
from octoprint.cli.common import LazyGroup

click.disable_unicode_literals_warning = True


# ~~ click context


class OctoPrintContext:
    """Custom context wrapping the standard options."""

    def __init__(self, configfile=None, basedir=None, verbosity=0, safe_mode=False):
        self.configfile = configfile
        self.basedir = basedir
        self.verbosity = verbosity
        self.safe_mode = safe_mode


pass_octoprint_ctx = click.make_pass_decorator(OctoPrintContext, ensure=True)
"""Decorator to pass in the :class:`OctoPrintContext` instance."""

# ~~ Basic CLI initialization for plugins


def init_platform_for_cli(ctx):
    """
    Performs a basic platform initialization for the CLI.

    Plugin implementations will be initialized, but only with a subset of the usual
    property injections:

       * _identifier and everything else parsed from metadata
       * _logger
       * _connectivity_checker
       * _environment_detector
       * _event_bus
       * _plugin_manager
       * _settings

    Returns: the same list of components as returned by ``init_platform``
    """

    from octoprint import (
        init_custom_events,
        init_platform,
        init_settings_plugin_config_migration_and_cleanup,
        init_webcam_compat_overlay,
    )
    from octoprint import octoprint_plugin_inject_factory as opif
    from octoprint import settings_plugin_inject_factory as spif

    components = init_platform(
        get_ctx_obj_option(ctx, "basedir", None),
        get_ctx_obj_option(ctx, "configfile", None),
        overlays=get_ctx_obj_option(ctx, "overlays", None),
        safe_mode=True,
        disable_color=get_ctx_obj_option(ctx, "no_color", False),
    )

    (
        settings,
        logger,
        safe_mode,
        event_manager,
        connectivity_checker,
        plugin_manager,
        environment_detector,
    ) = components

    init_custom_events(plugin_manager)
    octoprint_plugin_inject_factory = opif(
        settings,
        {
            "plugin_manager": plugin_manager,
            "event_bus": event_manager,
            "connectivity_checker": connectivity_checker,
            "environment_detector": environment_detector,
        },
    )
    settings_plugin_inject_factory = spif(settings)

    plugin_manager.implementation_inject_factories = [
        octoprint_plugin_inject_factory,
        settings_plugin_inject_factory,
    ]
    plugin_manager.initialize_implementations()

    init_settings_plugin_config_migration_and_cleanup(plugin_manager)
    init_webcam_compat_overlay(settings, plugin_manager)

    return components


# ~~ Custom click option to hide from help


class HiddenOption(click.Option):
    """Custom option sub class with empty help."""

    def get_help_record(self, ctx):
        pass


def hidden_option(*param_decls, **attrs):
    """Attaches a hidden option to the command.  All positional arguments are
    passed as parameter declarations to :class:`Option`; all keyword
    arguments are forwarded unchanged.  This is equivalent to creating an
    :class:`Option` instance manually and attaching it to the
    :attr:`Command.params` list.
    """

    import inspect

    from click.decorators import _param_memo

    def decorator(f):
        if "help" in attrs:
            attrs["help"] = inspect.cleandoc(attrs["help"])
        _param_memo(f, HiddenOption(param_decls, **attrs))
        return f

    return decorator


# ~~ helper for setting context options


def set_ctx_obj_option(ctx, param, value):
    """Helper for setting eager options on the context."""
    if ctx.obj is None:
        ctx.obj = OctoPrintContext()
    if value != param.default:
        setattr(ctx.obj, param.name, value)
    elif param.default is not None:
        setattr(ctx.obj, param.name, param.default)


# ~~ helper for retrieving context options


def get_ctx_obj_option(ctx, key, default, include_parents=True):
    if include_parents and hasattr(ctx, "parent") and ctx.parent:
        fallback = get_ctx_obj_option(ctx.parent, key, default)
    else:
        fallback = default
    return getattr(ctx.obj, key, fallback)


# ~~ helper for setting a lot of bulk options


def bulk_options(options):
    """
    Utility decorator to decorate a function with a list of click decorators.

    The provided list of ``options`` will be reversed to ensure correct
    processing order (inverse from what would be intuitive).
    """

    def decorator(f):
        options.reverse()
        for option in options:
            option(f)
        return f

    return decorator


# ~~ helper for setting --basedir, --config and --verbose options


def standard_options(hidden=False):
    """
    Decorator to add the standard options shared among all "octoprint" commands.

    If ``hidden`` is set to ``True``, the options will be available on the command but not
    listed in its help page.
    """

    factory = click.option
    if hidden:
        factory = hidden_option

    options = [
        factory(
            "--basedir",
            "-b",
            type=click.Path(),
            callback=set_ctx_obj_option,
            is_eager=True,
            expose_value=False,
            help="Specify the basedir to use for configs, uploads, timelapses etc.",
        ),
        factory(
            "--config",
            "-c",
            "configfile",
            type=click.Path(),
            callback=set_ctx_obj_option,
            is_eager=True,
            expose_value=False,
            help="Specify the config file to use.",
        ),
        factory(
            "--overlay",
            "overlays",
            type=click.Path(),
            callback=set_ctx_obj_option,
            is_eager=True,
            multiple=True,
            expose_value=False,
            help="Specify additional config overlays to use.",
        ),
        factory(
            "--verbose",
            "-v",
            "verbosity",
            count=True,
            callback=set_ctx_obj_option,
            is_eager=True,
            expose_value=False,
            help="Increase logging verbosity.",
        ),
        factory(
            "--safe",
            "safe_mode",
            is_flag=True,
            callback=set_ctx_obj_option,
            is_eager=True,
            expose_value=False,
            help="Enable safe mode; disables all third party plugins.",
        ),
        factory(
            "--no-color",
            "no_color",
            is_flag=True,
            callback=set_ctx_obj_option,
            is_eager=True,
            expose_value=False,
            help="Disable colored console output. Alternatively set the NO_COLOR environment variable to a value of 1.",
        ),
    ]

    return bulk_options(options)


# ~~ helper for settings legacy options we still have to support on "octoprint"

legacy_options = bulk_options(
    [
        hidden_option("--host", type=click.STRING, callback=set_ctx_obj_option),
        hidden_option("--port", type=click.INT, callback=set_ctx_obj_option),
        hidden_option("--logging", type=click.Path(), callback=set_ctx_obj_option),
        hidden_option("--debug", "-d", is_flag=True, callback=set_ctx_obj_option),
        hidden_option(
            "--daemon",
            type=click.Choice(["start", "stop", "restart"]),
            callback=set_ctx_obj_option,
        ),
        hidden_option(
            "--pid",
            type=click.Path(),
            default="/tmp/octoprint.pid",
            callback=set_ctx_obj_option,
        ),
        hidden_option(
            "--iknowwhatimdoing", "allow_root", is_flag=True, callback=set_ctx_obj_option
        ),
        hidden_option(
            "--ignore-blacklist",
            "ignore_blacklist",
            is_flag=True,
            callback=set_ctx_obj_option,
        ),
    ]
)
"""Legacy options available directly on the "octoprint" command in earlier versions.
   Kept available for reasons of backwards compatibility, but hidden from the
   generated help pages."""

# ~~ command groups from sub modules


@click.group()
def subcommands():
    pass


@subcommands.group(
    name="analysis", cls=LazyGroup, import_name="octoprint.cli.analysis:cli"
)
def analysis():
    """Analysis tools."""
    pass


@subcommands.group(name="client", cls=LazyGroup, import_name="octoprint.cli.client:cli")
def client():
    """Basic API client."""
    pass


@subcommands.group(name="config", cls=LazyGroup, import_name="octoprint.cli.config:cli")
def config():
    """Basic config manipulation."""
    pass


@subcommands.group(name="dev", cls=LazyGroup, import_name="octoprint.cli.dev:cli")
def dev():
    """Additional commands for development tasks."""
    pass


@subcommands.group(name="plugins", cls=LazyGroup, import_name="octoprint.cli.plugins:cli")
def plugins():
    """Additional commands provided by plugins."""
    pass


@subcommands.group(
    name="timelapse", cls=LazyGroup, import_name="octoprint.cli.timelapse:cli"
)
def timelapse():
    """Timelapse related commands."""
    pass


@subcommands.group(name="user", cls=LazyGroup, import_name="octoprint.cli.user:cli")
def user():
    """User management."""
    pass


# ~~ "octoprint" command, merges server_commands and plugin_commands groups

from .server import cli as server_commands  # noqa: E402
from .systeminfo import cli as systeminfo_commands  # noqa: E402


@click.group(
    name="octoprint",
    invoke_without_command=True,
    cls=click.CommandCollection,
    sources=[
        subcommands,
        server_commands,
        systeminfo_commands,
    ],
)
@standard_options()
@legacy_options
@click.version_option(version=octoprint.__version__, allow_from_autoenv=False)
@click.pass_context
def octo(ctx, **kwargs):
    if ctx.invoked_subcommand is None:
        # We have to support calling the octoprint command without any
        # sub commands to remain backwards compatible.
        #
        # But better print a message to inform people that they should
        # use the sub commands instead.

        def get_value(key):
            return get_ctx_obj_option(ctx, key, kwargs.get(key))

        daemon = get_value("daemon")

        if daemon:
            click.echo(
                'Daemon operation via "octoprint --daemon '
                'start|stop|restart" is deprecated, please use '
                '"octoprint daemon start|stop|restart" from now on'
            )

            if sys.platform == "win32" or sys.platform == "darwin":
                click.echo(
                    "Sorry, daemon mode is not supported under your operating system right now"
                )
            else:
                from octoprint.cli.server import daemon_command

                ctx.invoke(daemon_command, command=daemon, **kwargs)
        else:
            click.echo(
                'Starting the server via "octoprint" is deprecated, '
                'please use "octoprint serve" from now on.'
            )

            from octoprint.cli.server import serve_command

            ctx.invoke(serve_command, **kwargs)



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2017 The OctoPrint Project - Released under terms of the AGPLv3 License"

import base64
import logging

import requests

from . import ApiCheckError

BRANCH_HEAD_URL = (
    "https://api.bitbucket.org/2.0/repositories/{user}/{repo}/commit/{branch}"
)

logger = logging.getLogger(
    "octoprint.plugins.softwareupdate.version_checks.bitbucket_commit"
)


class BitbucketApiError(ApiCheckError):
    API = "Bitbucket API"


def check_bitbucket_api_response(logger, r, ok_codes=None):
    if ok_codes is None:
        ok_codes = [requests.codes.ok]

    if r.status_code not in ok_codes:
        try:
            data = r.json()
            message = data.get("message", "Unknown error")
        except Exception:
            message = "Not a valid JSON response"

        exc = BitbucketApiError(r.status_code, message)
        logger.error(exc.message)
        raise exc


def _get_latest_commit(user, repo, branch, api_user=None, api_password=None):
    from ..exceptions import NetworkError

    url = BRANCH_HEAD_URL.format(user=user, repo=repo, branch=branch)
    headers = {}
    if api_user is not None and api_password is not None:
        auth_value = base64.b64encode(
            b"{user}:{pw}".format(user=api_user, pw=api_password)
        )
        headers["authorization"] = f"Basic {auth_value}"

    try:
        r = requests.get(url, headers=headers, timeout=(3.05, 7))
    except requests.ConnectionError as exc:
        raise NetworkError(cause=exc)

    check_bitbucket_api_response(logger, r)

    reference = r.json()
    if "hash" not in reference:
        raise BitbucketApiError(r.status_code, "No commit hash found in response")

    return reference["hash"]


def get_latest(target, check, online=True, credentials=None, *args, **kwargs):
    from ..exceptions import ConfigurationInvalid

    if "user" not in check or "repo" not in check:
        raise ConfigurationInvalid(
            "Update configuration for %s of type bitbucket_commit needs all of user and repo"
            % target
        )

    branch = "master"
    if "branch" in check and check["branch"] is not None:
        branch = check["branch"]

    api_user = check.get("api_user")
    api_password = check.get("api_password")

    if api_user is None and api_password is None and credentials:
        api_user = credentials.get("bitbucket_user")
        api_password = credentials.get("bitbucket_password")

    current = check.get("current")

    information = {
        "local": {
            "name": "Commit {commit}".format(
                commit=current if current is not None else "unknown"
            ),
            "value": current,
        },
        "remote": {"name": "?", "value": "?"},
        "needs_online": not check.get("offline", False),
    }
    if not online and information["needs_online"]:
        return information, True

    remote_commit = _get_latest_commit(
        check["user"], check["repo"], branch, api_user, api_password
    )
    remote_name = f"Commit {remote_commit}" if remote_commit is not None else "-"

    information["remote"] = {"name": remote_name, "value": remote_commit}
    is_current = (
        current is not None and current == remote_commit
    ) or remote_commit is None

    logger.debug(f"Target: {target}, local: {current}, remote: {remote_commit}")

    return information, is_current



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

from enum import Enum
from typing import List

from octoprint.schema import BaseModel
from octoprint.vendor.with_attrs_docs import with_attrs_docs


class ColorEnum(str, Enum):
    red = "red"
    orange = "orange"
    yellow = "yellow"
    green = "green"
    blue = "blue"
    violet = "violet"
    default = "default"


@with_attrs_docs
class ComponentOrderConfig(BaseModel):
    navbar: List[str] = [
        "settings",
        "systemmenu",
        "plugin_announcements",
        "plugin_logging_seriallog",
        "plugin_logging_plugintimingslog",
        "plugin_pi_support",
        "login",
    ]
    """Order of navbar items."""

    sidebar: List[str] = [
        "plugin_firmware_check_warning",
        "plugin_firmware_check_info",
        "connection",
        "state",
        "files",
    ]
    """Order of sidebar items."""

    tab: List[str] = [
        "temperature",
        "control",
        "plugin_gcodeviewer",
        "terminal",
        "timelapse",
    ]
    """Order of tabs."""

    settings: List[str] = [
        "section_printer",
        "serial",
        "printerprofiles",
        "temperatures",
        "terminalfilters",
        "gcodescripts",
        "section_features",
        "features",
        "webcam",
        "accesscontrol",
        "plugin_gcodeviewer",
        "api",
        "plugin_appkeys",
        "section_octoprint",
        "server",
        "folders",
        "appearance",
        "plugin_logging",
        "plugin_pluginmanager",
        "plugin_softwareupdate",
        "plugin_announcements",
        "plugin_eventmanager",
        "plugin_backup",
        "plugin_tracking",
        "plugin_errortracking",
        "plugin_pi_support",
    ]
    """Order of settings."""

    usersettings: List[str] = ["access", "interface"]
    """Order of user settings."""

    wizard: List[str] = [
        "plugin_softwareupdate_update",
        "plugin_backup",
        "plugin_corewizard_acl",
        "plugin_corewizard_onlinecheck",
    ]
    """Order of wizards."""

    about: List[str] = [
        "about",
        "plugin_pi_support",
        "supporters",
        "authors",
        "changelog",
        "license",
        "thirdparty",
        "plugin_pluginmanager",
    ]
    """Order of about dialog items."""

    generic: List[str] = []
    """Order of generic items."""


@with_attrs_docs
class ComponentDisabledConfig(BaseModel):
    navbar: List[str] = []
    """Disabled navbar items."""
    sidebar: List[str] = []
    """Disabled sidebar items."""
    tab: List[str] = []
    """Disabled tabs."""
    settings: List[str] = []
    """Disabled settings."""
    usersettings: List[str] = []
    """Disabled user settings."""
    wizard: List[str] = []
    """Disabled wizards."""
    about: List[str] = []
    """Disabled about dialog items."""
    generic: List[str] = []
    """Disabled generic items."""


@with_attrs_docs
class ComponentConfig(BaseModel):
    order: ComponentOrderConfig = ComponentOrderConfig()
    """Defines the order of the components within their respective containers."""

    disabled: ComponentDisabledConfig = ComponentDisabledConfig()
    """Disabled components per container. If a component is included here it will not be included in OctoPrint's UI at all. Note that this might mean that critical functionality will not be available if no replacement is registered."""


@with_attrs_docs
class AppearanceConfig(BaseModel):
    name: str = ""
    """Use this to give your OctoPrint instance a name. It will be displayed in the title bar (as "<Name> [OctoPrint]") and in the navigation bar (as "OctoPrint: <>")"""

    color: ColorEnum = ColorEnum.default
    """Use this to color the navigation bar."""

    colorTransparent: bool = False
    """Makes the color of the navigation bar "transparent". In case your printer uses acrylic for its frame 😉."""

    colorIcon: bool = True

    defaultLanguage: str = "_default"
    """Default language of OctoPrint. If left unset OctoPrint will try to match up available languages with the user's browser settings."""

    showFahrenheitAlso: bool = False

    fuzzyTimes: bool = True

    closeModalsWithClick: bool = True

    showInternalFilename: bool = True
    """Show the internal filename in the files sidebar, if necessary."""

    components: ComponentConfig = ComponentConfig()
    """Configures the order and availability of the UI components."""



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

from typing import List, Union

from octoprint.schema import BaseModel
from octoprint.vendor.with_attrs_docs import with_attrs_docs

from .access_control import AccessControlConfig
from .api import ApiConfig
from .appearance import AppearanceConfig
from .controls import ContainerConfig, ControlConfig
from .devel import DevelConfig
from .estimation import EstimationConfig
from .events import EventsConfig
from .feature import FeatureConfig
from .folder import FolderConfig
from .gcode_analysis import GcodeAnalysisConfig
from .plugins import PluginsConfig
from .printer_parameters import PrinterParametersConfig
from .printer_profiles import PrinterProfilesConfig
from .scripts import ScriptsConfig
from .serial import SerialConfig
from .server import ServerConfig
from .slicing import SlicingConfig
from .system import SystemConfig
from .temperature import TemperatureConfig
from .terminalfilters import DEFAULT_TERMINAL_FILTERS, TerminalFilterEntry
from .webcam import WebcamConfig


@with_attrs_docs
class Config(BaseModel):
    accessControl: AccessControlConfig = AccessControlConfig()
    api: ApiConfig = ApiConfig()
    appearance: AppearanceConfig = AppearanceConfig()
    controls: List[Union[ControlConfig, ContainerConfig]] = []
    devel: DevelConfig = DevelConfig()
    estimation: EstimationConfig = EstimationConfig()
    events: EventsConfig = EventsConfig()
    feature: FeatureConfig = FeatureConfig()
    folder: FolderConfig = FolderConfig()
    gcodeAnalysis: GcodeAnalysisConfig = GcodeAnalysisConfig()
    plugins: PluginsConfig = PluginsConfig()
    printerParameters: PrinterParametersConfig = PrinterParametersConfig()
    printerProfiles: PrinterProfilesConfig = PrinterProfilesConfig()
    scripts: ScriptsConfig = ScriptsConfig()
    serial: SerialConfig = SerialConfig()
    server: ServerConfig = ServerConfig()
    slicing: SlicingConfig = SlicingConfig()
    system: SystemConfig = SystemConfig()
    temperature: TemperatureConfig = TemperatureConfig()
    terminalFilters: List[TerminalFilterEntry] = DEFAULT_TERMINAL_FILTERS
    webcam: WebcamConfig = WebcamConfig()



"""
This module bundles platform specific flags and implementations.
"""

__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2017 The OctoPrint Project - Released under terms of the AGPLv3 License"

import sys

try:
    import fcntl
except ImportError:
    fcntl = None

# set_close_exec

if fcntl is not None and hasattr(fcntl, "FD_CLOEXEC"):

    def set_close_exec(handle):
        """Set ``close_exec`` flag on handle, if supported by the OS."""
        flags = fcntl.fcntl(handle, fcntl.F_GETFD)
        flags |= fcntl.FD_CLOEXEC
        fcntl.fcntl(handle, fcntl.F_SETFD, flags)

elif sys.platform == "win32":

    def set_close_exec(handle):
        """Set ``close_exec`` flag on handle, if supported by the OS."""
        import ctypes
        import ctypes.wintypes

        # see https://msdn.microsoft.com/en-us/library/ms724935(v=vs.85).aspx
        SetHandleInformation = ctypes.windll.kernel32.SetHandleInformation
        SetHandleInformation.argtypes = (
            ctypes.wintypes.HANDLE,
            ctypes.wintypes.DWORD,
            ctypes.wintypes.DWORD,
        )
        SetHandleInformation.restype = ctypes.c_bool

        HANDLE_FLAG_INHERIT = 0x00000001

        result = SetHandleInformation(handle, HANDLE_FLAG_INHERIT, 0)
        if not result:
            raise ctypes.GetLastError()

else:

    def set_close_exec(handle):
        """Set ``close_exec`` flag on handle, if supported by the OS."""
        # no-op
        pass


# default close_fds settings
CLOSE_FDS = True
"""
Default setting for close_fds parameter to Popen/sarge.run.

Set ``close_fds`` on every sub process to this to ensure file handlers will be closed
on child processes on platforms.
"""

# current os

_OPERATING_SYSTEMS = {
    "windows": ["win32"],
    "linux": lambda x: x.startswith("linux"),
    "macos": ["darwin"],
    "freebsd": lambda x: x.startswith("freebsd"),
}
OPERATING_SYSTEM_UNMAPPED = "unmapped"


def get_os():
    """
    Returns a canonical OS identifier.

    Currently the following OS are recognized: ``win32``, ``linux`` (``sys.platform`` = ``linux*``),
    ``macos`` (``sys.platform`` = ``darwin``) and ``freebsd`` (``sys.platform`` = ``freebsd*``).

    Returns:
            (str) mapped OS identifier
    """
    for identifier, platforms in _OPERATING_SYSTEMS.items():
        if (callable(platforms) and platforms(sys.platform)) or (
            isinstance(platforms, list) and sys.platform in platforms
        ):
            return identifier
    else:
        return OPERATING_SYSTEM_UNMAPPED


def is_os_compatible(compatibility_entries, current_os=None):
    """
    Tests if the ``current_os`` or ``sys.platform`` are blacklisted or whitelisted in ``compatibility_entries``

    Returns:
            (bool) True if the os is compatible, False otherwise
    """
    if len(compatibility_entries) == 0:
        # shortcut - no compatibility info means we are compatible
        return True

    if current_os is None:
        current_os = get_os()

    negative_entries = list(
        map(lambda x: x[1:], filter(lambda x: x.startswith("!"), compatibility_entries))
    )
    positive_entries = list(
        filter(lambda x: not x.startswith("!"), compatibility_entries)
    )

    negative_match = False
    if negative_entries:
        # check if we are blacklisted
        negative_match = current_os in negative_entries or any(
            map(lambda x: sys.platform.startswith(x), negative_entries)
        )

    positive_match = True
    if positive_entries:
        # check if we are whitelisted
        positive_match = current_os in positive_entries or any(
            map(lambda x: sys.platform.startswith(x), positive_entries)
        )

    return positive_match and not negative_match
